#!/usr/bin/env python
# -*- coding: utf-8 -*-

from copy import deepcopy
import warnings


def deprecate(old, new=None, version=None, ref=None):
    """
    Decorator to raise deprecation warning.

    Args:
        old (str): description of the old method/function
        new (str or None): description of the new method/function
        version (str or None): version number, like 2.7.3-alpha
        ref (str or None): reference URL of the new method/function
    """
    def _deprecate(func):
        def wrapper(*args, **kwargs):
            version_str = "." if version is None else f", version >= {version}."
            message = "" if ref is None else f" Refer to {ref}."
            if new is None:
                comment = f"{old} was deprecated{version_str}{message}"
            else:
                comment = f"Please use {new} rather than {old}{version_str}{message}"
            warnings.warn(
                comment,
                DeprecationWarning,
                stacklevel=2
            )
            return func(*args, **kwargs)
        return wrapper
    return _deprecate


def experimental(name, version):
    """
    Decorator to raise ExperimentalWarning because the method/function is experimental.

    Args:
        name (str): description of the method/function
        version (str): version number, like 2.7.3-alpha
    """
    def _experimental(func):
        def wrapper(*args, **kwargs):
            comment = f"{name} can be used from {version}, but this is experimental." \
                "Its name and arguments may be changed later."
            warnings.warn(
                comment,
                ExperimentalWarning,
                stacklevel=2
            )
            return func(*args, **kwargs)
        return wrapper
    return _experimental


class _BaseWarning(Warning):
    """Basic class of warning.
    """
    pass


class ExperimentalWarning(Warning):
    """Class to explain the method/function is experimental and its name,
    features and arguments may changed later.
    """
    pass


class _BaseException(Exception):
    """Basic class of exception.

    Args:
        message (str): main message of error, should be set in child classes
        details (str or None): details of error
    """

    def __init__(self, message, details=None):
        self._message = str(message)
        self._details = "" if details is None else f" {details}."

    def __str__(self):
        return f"{self._message}. {self._details}"


class AlreadyCalledError(_BaseException):
    """Error when a method has already been called and cannot be called any more.

    Args:
        name (str): the name of the method
        details (str or None): details of error
    """

    def __init__(self, name, details=None):
        message = f"{name} has already been called and cannot be called any more"
        super().__init__(message=message, details=details)


class NotIncludedError(_BaseException):
    """Error when a necessary key was not included in a container.

    Args:
        key_name (str): key name
        container_name (str): name of the container
        details (str or None): details of error
    """

    def __init__(self, key_name, container_name, details=None):
        message = f"'{key_name}' was not included in the '{container_name}'"
        super().__init__(message=message, details=details)


class NAFoundError(_BaseException):
    """Error when NA values are included un-expectedly.

    Args:
        name (str): name of the target
        value (str or None): value of the target
        details (str or None): details of error
    """

    def __init__(self, name, value=None, details=None):
        message = f"'{name}' has NA(s) un-expectedly"
        if value is not None:
            message += f", '{value}'"
        super().__init__(message=message, details=details)


class NotEnoughDataError(_BaseException):
    """Error when we do not have enough data for analysis.

    Args:
        name (str): name of the target
        value (str): value of the target
        required_n (int): required number of records
        details (str or None): details of error
    """

    def __init__(self, name, value, required_n, details=None):
        message = f"We need more than {required_n} records, but '{name}' has only {len(value)} records at this time"
        super().__init__(message=message, details=details)


class UnExpectedNoneError(_BaseException):
    """Error when a value is None un-expectedly.

    Args:
        name (str): name of the target
        details (str or None): details of error
    """

    def __init__(self, name, details=None):
        message = f"'{name}' is None un-expectedly"
        super().__init__(message=message, details=details)


class NotNoneError(_BaseException):
    """Error when a value must be None but not None un-expectedly.

    Args:
        name (str): name of the target
        value (str): value of the target
        details (str or None): details of error
    """

    def __init__(self, name, value, details=None):
        message = f"'{name}' must be None, but has value '{value}'"
        super().__init__(message=message, details=details)


class UnExecutedError(_BaseException):
    """
    Error when we have unexecuted methods that we need to run in advance.

    Args:
        name (str): method name to run in advance
        details (str or None): details of error
    """

    def __init__(self, name, details=None):
        message = f"Please execute {name} in advance"
        super().__init__(message=message, details=details)


class NotRegisteredError(UnExecutedError):
    """
    Error when no records have been registered yet.
    """
    pass


class NotRegisteredMainError(UnExecutedError):
    """
    Error when main datasets were not registered.
    """
    pass


class NotRegisteredExtraError(UnExecutedError):
    """
    Error when extra datasets were not registered.
    """
    pass


class NotSubclassError(_BaseException):
    """Error when an object is not a subclass of the parent class un-expectedly.

    Args:
        name (str): name of the target
        target (object): target object
        parent (object): expected parent class
        details (str or None): details of error
    """

    def __init__(self, name, target, parent, details=None):
        message = f"'{name}' must be a sub-class of {parent}, but {type(target)} was applied"
        super().__init__(message=message, details=details)


class UnExpectedTypeError(_BaseException):
    """Error when an object cannot be converted to an instance un-expectedly.

    Args:
        name (str): name of the target
        target (object): target object
        expected (object): expected type
        details (str or None): details of error
    """

    def __init__(self, name, target, expected, details=None):
        message = f"We could not convert '{name}' to an instance of {expected} because that of {type(target)} was applied"
        super().__init__(message=message, details=details)


class EmptyError(_BaseException):
    """Error when the dataframe is empty un-expectedly.

    Args:
        name (str): name of the target
        details (str or None): details of error
    """

    def __init__(self, name, details=None):
        message = f"'Empty dataframe/series was applied as {name}' un-expectedly"
        super().__init__(message=message, details=details)


class UnExpectedValueRangeError(_BaseException):
    """Error when the value is out of value range.

    Args:
        name (str): name of the target
        target (object): target object
        value_range (tuple(int or None, int or None)): value range, None means un-specified
        details (str or None): details of error
    """

    def __init__(self, name, target, value_range, details=None):
        _min, _max = value_range
        if _min is None:
            s = "is not in the expected value range" if _max is None else f"must be under or equal to {_max}"
        else:
            s = f"must be over or equal to {_min}" if _max is None else f"is not in the expected value range ({_min}, {_max})"
        message = f"'{name}' {s}, but {target} was applied"
        super().__init__(message=message, details=details)


class UnExpectedValueError(_BaseException):
    """
    Error when unexpected value was applied as the value of an argument.

    Args:
        name (str): argument name
        value (object): value user applied
        candidates (list[object]): candidates of the argument
        details (str or None): details of error
    """

    def __init__(self, name, value, candidates, details=None):
        c_str = ", ".join(candidates)
        message = f"'{name}' must be selected from [{c_str}], but {value} was applied"
        super().__init__(message=message, details=details)


class UnExpectedLengthError(_BaseException):
    """
    Error when a sequence has un-expended length.

    Args:
        name (str): argument name
        value (object): value user applied
        length (int): length of the sequence
        details (str or None): details of error
    """

    def __init__(self, name, value, length, details=None):
        message = f"The length of '{name}' must be {length}, but {len(value)} was applied"
        super().__init__(message=message, details=details)


class SubsetNotFoundError(_BaseException):
    """Error when subset was failed with specified arguments.

    Args:
        geo (tuple(list[str] or tuple(str) or str) or str or None): location names to filter or None
        country (str): country name
        country_alias (str or None): country name used in the dataset
        province (str or None): province name
        start_date (str or None): start date, like 22Jan2020
        end_date (str or None): end date, like 01Feb2020
        date (str or None): specified date, like 22Jan2020
        details (str or None): details of error
    """

    def __init__(self, geo=None, country=None, country_alias=None, province=None,
                 start_date=None, end_date=None, date=None, details=None):
        self.area = self._area(geo, country, country_alias, province)
        self.date = self._date(start_date, end_date, date)
        message = f"No records in {self.area}{self.date} were found."
        super().__init__(message=message, details=details)

    @staticmethod
    def _area(geo, country, country_alias, province):
        """
        Error when subset was failed with specified arguments.

        Args:
            geo (tuple(list[str] or tuple(str) or str) or str or None): location names to filter or None
            country (str): country name
            country_alias (str or None): country name used in the dataset
            province (str or None): province name

        Returns:
            str: area name
        """
        if geo is None and country is None:
            return "the world"
        if geo is not None:
            geo_converted = deepcopy(geo)
        elif province is None:
            geo_converted = (country if country_alias is None else f"{country} ({country_alias})",)
        else:
            geo_converted = (country if country_alias is None else f"{country} ({country_alias})", province)
        names = [
            info if isinstance(info, str) else "_".join(list(info))
            for info in ([geo_converted] if isinstance(geo_converted, str) else geo_converted)]
        return "/".join(names[::-1])

    @staticmethod
    def _date(start_date, end_date, date):
        """
        Error when subset was failed with specified arguments.

        Args:
            start_date (str or None): start date, like 22Jan2020
            end_date (str or None): end date, like 01Feb2020
            date (str or None): specified date, like 22Jan2020
        """
        if date is not None:
            return f" on {date}"
        start_str = "" if start_date is None else f" from {start_date}"
        end_str = "" if end_date is None else f" to {end_date}"
        return f"{start_str}{end_str}"


class ScenarioNotFoundError(_BaseException):
    """Error when unregistered scenario name was specified.

    Args:
        name (str): scenario name
        details (str or None): details of error
    """

    def __init__(self, name, details=None):
        message = f"{name} scenario is not registered."
        super().__init__(message=message, details=details)


class PCRIncorrectPreconditionError(_BaseException):
    """Error when checking preconditions in the PCR data.

    Args:
        country (str): country name
        province (str or None): province name
        details (str or None): details of error
    """

    def __init__(self, country, province=None, details=None):
        self.area = self._area(country, province)
        message = f"The dataset of {self.area} has too many missing values"
        super().__init__(message=message, details=details)

    @staticmethod
    def _area(country, province):
        """
        Error when PCR preconditions failed with specified arguments.

        Args:
            country (str): country name
            province (str or None): province name

        Returns:
            str: area name
        """
        if province == "-":
            province = None
        country_str = "" if province else f" in country {country}"
        province_str = "" if province is None else f" in province {province}"
        return f"{province_str}{country_str}"


class NotInteractiveError(_BaseException):
    """Error when interactive shell is not used but forced to use it.

    Args:
        details (str or None): details of error
    """

    def __init__(self, details=None):
        message = "Interactive shell is not used."
        super().__init__(message=message, details=details)


class UnExpectedReturnValueError(_BaseException):
    """Error when unexpected value was returned.

    Args:
        name (str): argument name
        value (object): value user applied or None (will not be shown)
        plural (bool): whether plural or not
        details (str or None): details of error
    """

    def __init__(self, name, value, plural=False, details=None):
        self.value = "" if value is None else f" ({value})"
        self.s = "s" if plural else ""
        self.be = "were" if plural else "was"
        message = f"Un-expected value{self.s}{self.value} {self.be} returned as {name}"
        super().__init__(message=message, details=details)


class DBLockedError(_BaseException):
    """Error when a database has been locked not as expected.

    Args:
        name (str): database name
        details (str or None): details of error
    """

    def __init__(self, name, details=None):
        message = f"{name} should NOT be locked, but locked"
        super().__init__(message=message, details=details)


class NotDBLockedError(_BaseException):
    """Error when a database has NOT been locked not as expected.

    Args:
        name (str): database name
        details (str or None): details of error
    """

    def __init__(self, name, details=None):
        message = f"{name} should be locked, but NOT locked"
        super().__init__(message=message, details=details)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import contextlib
import pandas as pd
from covsirphy.util.validator import Validator
from covsirphy.util.term import Term


class _LayerAdjuster(Term):
    """Class to adjust location layers of time-series data.

    Args:
        layers (list[str] or None): list of layers of geographic information or None (["ISO3", "Province", "City"])
        country (str or None): layer name of countries or None (countries are not included in the layers)
        date (str): column name of observation dates
        verbose (int): level of verbosity of stdout

    Raises:
        ValueError: @layers has duplicates

    Note:
        Country level data specified with @country will be stored with ISO3 codes.

    Note:
        If @verbose is 0, no descriptions will be shown.
        If @verbose is 1 or larger, details of layer adjustment will be shown.
    """
    # Internal term
    _ID = "Location_ID"

    def __init__(self, layers=None, country="ISO3", date="Date", verbose=1):
        # Countries will be specified with ISO3 codes and this requires conversion
        self._country = None if country is None else str(country)
        # Layers of location information
        self._layers = Validator(layers or [self._country, self.PROVINCE, self.CITY], "layers").sequence()
        if len(set(self._layers)) != len(self._layers):
            raise ValueError(f"@layer has duplicates, {self._layers}")
        # Date column
        self._date = str(date)
        # Verbosity
        self._verbose = Validator(verbose, "verbose").int(value_range=(0, None))
        # Location data
        self._loc_df = pd.DataFrame(columns=[self._ID, *self._layers])
        # All available data
        self._rec_df = pd.DataFrame(columns=[self._ID, self._date])
        # Citations
        self._citation_dict = {}

    def all(self, variables=None):
        """Return all available data.

        Args:
            variables (list[str] or None): list of variables to collect or None (all available variables)

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - columns defined by covsirphy.LayerAdjuster(layers)
                    - Date (pandas.Timestamp): observation dates
                    - columns defined by @variables
        """
        identifiers = [*self._layers, self._date]
        df = self._loc_df.merge(self._rec_df, how="right", on=self._ID).drop(self._ID, axis=1)
        df = df.sort_values(identifiers, ignore_index=True)
        df = df.loc[:, [*identifiers, *sorted(set(df.columns) - set(identifiers), key=df.columns.tolist().index)]]
        if variables is None:
            return df
        all_variables = df.columns.tolist()
        sel_variables = Validator(variables, "variables").sequence(candidates=set(all_variables) - set(identifiers))
        return df.loc[:, [*self._layers, self._date, *sel_variables]]

    def citations(self, variables=None):
        """
        Return citation list of the secondary data sources.

        Args:
            variables (list[str] or None): list of variables to collect or None (all available variables)

        Returns:
            list[str]: citation list
        """
        all_columns = [col for col in self._rec_df.columns if col not in (self._ID, self._date)]
        columns = Validator(variables, "variables").sequence(default=all_columns, candidates=all_columns)
        return Validator([v for k, v in self._citation_dict.items() if k in columns]).sequence(flatten=True, unique=True)

    def register(self, data, layers=None, date="Date", variables=None, citations=None, convert_iso3=True, **kwargs):
        """Register new data.

        Args:
            data (pandas.DataFrame): new data
                Index
                    reset index
                Columns
                    - columns defined by @data_layers
                    - columns defined by @date
                    - columns defined by @variables
            layers (list[str] or None): layers of the data or None (as the same _LayerAdjuster(layer))
            date (str): column name of observation dates of the data
            variables (list[str] or None): list of variables to add or None (all available columns)
            citations (list[str] or str or None): citations of the dataset or None (["my own dataset"])
            convert_iso3 (bool): whether convert country names to ISO3 codes or not
            **kwargs: keyword arguments of pandas.to_datetime() including "dayfirst (bool): whether date format is DD/MM or not"

        Raises:
            ValueError: @data_layers has duplicates

        Returns:
            covsirphy.LayerAdjuster: self
        """
        data_layers = Validator(layers, "layers").sequence(default=self._layers)
        if len(set(data_layers)) != len(data_layers):
            raise ValueError(f"@layer has duplicates, {data_layers}")
        df = Validator(data, "data").dataframe(columns=[*data_layers, date, *(variables or [])])
        # Convert date type
        df.rename(columns={date: self._date}, inplace=True)
        df[self._date] = pd.to_datetime(df[self._date], **kwargs).dt.round("D")
        with contextlib.suppress(TypeError):
            df[self.DATE] = df[self.DATE].dt.tz_convert(None)
        # Convert country names to ISO3 codes
        if convert_iso3 and self._country is not None and self._country in df:
            df.loc[:, self._country] = self._to_iso3(df[self._country])
        # Prepare necessary layers and fill in None with "NA"
        if data_layers is not None and data_layers != self._layers:
            df = self._prepare_layers(df, data_layers=data_layers)
        df.loc[:, self._layers] = df[self._layers].astype("string").fillna(self.NA)
        # Locations
        loc_df = df.loc[:, self._layers]
        loc_df = pd.concat([self._loc_df, loc_df], axis=0, ignore_index=True)
        loc_df.drop_duplicates(subset=self._layers, keep="first", ignore_index=True, inplace=True)
        loc_df.loc[loc_df[self._ID].isna(), self._ID] = f"id{len(loc_df)}-" + \
            loc_df[loc_df[self._ID].isna()].index.astype("str")
        self._loc_df = loc_df.reset_index()[[self._ID, *self._layers]]
        # Records
        df = df.merge(self._loc_df, how="left", on=self._layers).drop(self._layers, axis=1)
        if variables is not None:
            columns = [self._ID, self._date, *Validator(variables, "variables").sequence()]
            df = df.loc[:, columns]
        rec_df = self._rec_df.reindex(columns=list(set(self._rec_df.columns) | set(df.columns)))
        rec_df = rec_df.set_index([self._ID, self._date]).combine_first(df.set_index([self._ID, self._date]))
        self._rec_df = rec_df.reset_index()
        # Citations
        new_citations = Validator(
            [citations] if isinstance(citations, str) else (citations or ["my own dataset"]), "citations").sequence()
        citation_dict = {col: self._citation_dict.get(col, []) + new_citations for col in variables or df.columns}
        self._citation_dict.update(citation_dict)
        return self

    def _prepare_layers(self, data, data_layers):
        """Prepare necessary layers, adding NAs and renaming layers.

        Args:
            data (pandas.DataFrame): local dataset
                Index
                    reset index
                Columns
                    - columns defined by @data_layers
                    - the other columns
            data_layers (list[str]): layers of the data

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - columns defined by LayerAdjuster(layers)
                    - the other columns
        """
        df = data.copy()
        expected, actual = self._align_layers(data_layers)
        # Adjust layer names and records
        for (layer, data_layer) in zip(expected, actual):
            if data_layer is None:
                df.loc[:, layer] = self.NA
                self._print_v0(f"\t[INFO] New layer '{layer}' was added to the data with NAs.")
            elif layer is None:
                if data_layer == actual[-1]:
                    df.loc[df[data_layer] != self.NA, self._layers] = self.NA
                    self._print_v0(f"\t[INFO] Records which has actual values at '{data_layer}' layer were disabled.")
                df = df.drop(data_layer, axis=1)
                self._print_v0(f"\t[INFO] '{data_layer}' layer was removed.")
            elif layer != data_layer:
                df.rename(columns={data_layer: layer}, inplace=True)
                self._print_v0(f"\t[INFO] '{data_layer}' layer was renamed to {layer}.")
        return df.reset_index(drop=True)

    def _align_layers(self, data_layers):
        """Perform sequence alignment of the layers of new data with the layers defined by LayerAdjuster(layers).

        Args:
            data_layers (list[str]): layers of the data

        Returns:
            tuple(list[str], list[str]): list of aligned layers, that of defined and that of new data

        Note:
            Example of sequence alignment: [A, B, C], [A, C] -> [A, B, C], [A, None, C]
        """
        expected, actual = [], []
        for layer in self._layers:
            current = [data_layer for data_layer in data_layers if data_layer not in actual]
            if layer in current:
                new = current[:current.index(layer) + 1]
                expected.extend([None for _ in range(len(new) - 1)] + [layer])
                actual.extend(new)
            else:
                expected.append(layer)
                actual.append(None)
        not_included = [data_layer for data_layer in data_layers if data_layer not in actual]
        expected += [None for _ in range(len(not_included))]
        actual += not_included
        for (i, (e, a)) in enumerate(zip(expected[:], actual[:]), start=1):
            if i == len(expected):
                break
            if e is not None and expected[i] is None and a is None and actual[i] is not None:
                expected[i - 1:i + 1] = [e]
                actual[i - 1:i + 1] = [actual[i]]
        return expected, actual

    def _print_v0(self, sentence):
        """Stdout the sentence when the verbosity level was set to 1 or higher.

        Args:
            sentence (str): the sentence to stdout
        """
        if self._verbose:
            print(sentence)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from datetime import timedelta
import numpy as np
import pandas as pd
from covsirphy.util.argument import find_args
from covsirphy.util.error import deprecate, SubsetNotFoundError
from covsirphy.util.validator import Validator
from covsirphy._deprecated.jhu_data import JHUData
from covsirphy._deprecated._mbase import ModelBase
from covsirphy._deprecated.ode_handler import ODEHandler


class ExampleData(JHUData):
    """
    Deprecated. Example dataset as a child class of JHUData.

    Args:
        clean_df (pandas.DataFrame or None): cleaned data

            Index
                - reset index
            Columns
                - Date (pd.Timestamp): Observation date
                - Country (pandas.Category): country/region name
                - Province (pandas.Category): province/prefecture/state name
                - Confirmed (int): the number of confirmed cases
                - Infected (int): the number of currently infected cases
                - Fatal (int): the number of fatal cases
                - Recovered (int): the number of recovered cases

        tau (int): tau value [min]
        start_date (str): start date, like 22Jan2020
    """

    @deprecate(old="ExampleData()", new="Dynamics.from_sample()", version="2.24.0-kappa")
    def __init__(self, clean_df=None, tau=1440, start_date="22Jan2020"):
        if clean_df is None:
            clean_df = pd.DataFrame(columns=self.COLUMNS)
        clean_df = self._ensure_dataframe(clean_df, name="clean_df", columns=self.COLUMNS)
        variables = [self.C, self.CI, self.F, self.R, self.N]
        self._raw_cols = [self.DATE, self.ISO3, self.COUNTRY, self.PROVINCE, *variables]
        self._subset_cols = [self.DATE, *variables]
        self._raw = clean_df.copy()
        self._cleaned_df = clean_df.copy()
        self._citation = str()
        self._tau = Validator(tau, "tau").tau(default=None)
        self._start = Validator(start_date, "start_date").date()
        self._population = None
        self._recovery_period = None

    @staticmethod
    def _ensure_dataframe(target, name="df", time_index=False, columns=None, empty_ok=True):
        """
        Ensure the dataframe has the columns.

        Args:
            target (pandas.DataFrame): the dataframe to ensure
            name (str): argument name of the dataframe
            time_index (bool): if True, the dataframe must has DatetimeIndex
            columns (list[str] or None): the columns the dataframe must have
            empty_ok (bool): whether give permission to empty dataframe or not

        Returns:
            pandas.DataFrame:
                Index
                    as-is
                Columns:
                    columns specified with @columns or all columns of @target (when @columns is None)
        """
        if not isinstance(target, pd.DataFrame):
            raise TypeError(f"@{name} must be a instance of (pandas.DataFrame).")
        df = target.copy()
        if time_index and (not isinstance(df.index, pd.DatetimeIndex)):
            raise TypeError(f"Index of @{name} must be <pd.DatetimeIndex>.")
        if not empty_ok and target.empty:
            raise ValueError(f"@{name} must not be a empty dataframe.")
        if columns is None:
            return df
        if not set(columns).issubset(df.columns):
            cols_str = ", ".join(col for col in columns if col not in df.columns)
            included = ", ".join(df.columns.tolist())
            s1 = f"Expected columns were not included in {name} with {included}."
            raise KeyError(f"{s1} {cols_str} must be included.")
        return df.loc[:, columns]

    def ensure_country_name(self, country, errors="raise"):
        """
        Ensure that the country name is correct.

        Args:
            country (str): country name
            errors (str): 'raise' or 'coerce'

        Returns:
            str: country name

        Raises:
            SubsetNotFoundError: no records were found for the country and @errors is 'raise'
        """
        df = self._cleaned_df.copy()
        self._ensure_dataframe(df, name="the cleaned dataset", columns=[self.COUNTRY])
        selectable_set = set(df[self.COUNTRY].unique())
        if country in selectable_set:
            return country
        if errors == "raise":
            raise SubsetNotFoundError(country=country)

    def _model_to_area(self, model=None, country=None, province=None):
        """
        If country is None and model is not None, model name will returned as country name.

        Args:
            model (cs.ModelBase or None): the first ODE model
            country (str or None): country name
            province (str or None): province name

        Raises:
            ValueError: both of country and model are None

        Returns:
            tuple(str, str): country name and province name
        """
        province = province or self.NA
        if country is not None:
            return (country, province)
        if model is None:
            raise ValueError("@model or @country must be specified.")
        model = Validator(model, "model").subclass(ModelBase)
        return (model.NAME, province)

    def add(self, model, country=None, province=None, **kwargs):
        """
        Add example data.
        If the country has been registered,
        the start date will be the next data of the registered records.

        Args:
            model (cs.ModelBase): the first ODE model
            country (str or None): country name
            province (str or None): province name
            kwargs: the other keyword arguments of model.EXAMPLE

        Note:
            If country is None, the name of the model will be used.
            If province is None, '-' will be used.
        """
        # Arguments
        model = Validator(model, "model").subclass(ModelBase)
        arg_dict = model.EXAMPLE.copy()
        arg_dict.update(kwargs)
        self._population = arg_dict["population"]
        # Area
        country, province = self._model_to_area(model=model, country=country, province=province)
        # Start date and y0 values
        df = self._cleaned_df.copy()
        df = df.loc[(df[self.COUNTRY] == country) & (df[self.PROVINCE] == province), :]
        if df.empty:
            start = self._start
        else:
            start = df.loc[df.index[-1], self.DATE]
            df[self.S] = self._population - df[self.C]
            df = model.convert(df, tau=None)
            arg_dict[self.Y0_DICT] = {k: df.loc[df.index[0], k] for k in model.VARIABLES}
        # Simulation
        end = start + timedelta(days=int(arg_dict[self.STEP_N] * self._tau / 1440))
        handler = ODEHandler(model, start, tau=self._tau)
        handler.add(end, param_dict=arg_dict[self.PARAM_DICT], y0_dict=arg_dict[self.Y0_DICT])
        restored_df = handler.simulate()
        # JHU-type records
        restored_df[self.ISO3] = country
        restored_df[self.COUNTRY] = country
        restored_df[self.PROVINCE] = province
        restored_df[self.C] = restored_df[[self.CI, self.F, self.R]].sum(axis=1)
        restored_df[self.N] = self._population
        selected_df = restored_df.loc[:, self._raw_cols]
        cleaned_df = pd.concat([self._cleaned_df, selected_df], axis=0, ignore_index=True)
        for col in self.AREA_COLUMNS:
            cleaned_df[col] = cleaned_df[col].astype("category")
        for col in self.VALUE_COLUMNS:
            cleaned_df[col] = cleaned_df[col].astype(np.int64)
        self._cleaned_df = cleaned_df.copy()

    def specialized(self, model=None, country=None, province=None):
        """
        Return dimensional records with model variables.

        Args:
            model (cs.ModelBase or None): the first ODE model
            country (str or None): country name
            province (str or None): province name

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - Date (pd.Timestamp): Observation date
                    - (int) variables of the model

        Note:
            If country is None, the name of the model will be used.
            If province is None, '-' will be used.
        """
        restored_df = self.subset(model=model, country=country, province=province)
        restored_df[self.S] = self._population - restored_df[self.C]
        return model.convert(restored_df, tau=None).reset_index()

    def non_dim(self, model=None, country=None, province=None):
        """
        Return non-dimensional data.

        Args:
            model (cs.ModelBase or None): the first ODE model
            country (str or None): country name
            province (str or None): province name

        Returns:
            pandas.DataFrame:
                Index
                    t: Dates divided by tau value (time steps)
                Columns
                    - (int) variables of the model

        Note:
            If country is None, the name of the model will be used.
            If province is None, '-' will be used.
        """
        restored_df = self.subset(model=model, country=country, province=province)
        restored_df[self.S] = self._population - restored_df[self.C]
        df = model.convert(restored_df, tau=self._tau)
        return df.apply(lambda x: x / x.sum(), axis=1)

    def subset(self, model=None, country=None, province=None, **kwargs):
        """
        Return the subset of dataset.

        Args:
            model (cs.ModelBase or None): the first ODE model
            country (str or None): country name
            province (str or None): province name
            kwargs: other keyword arguments of JHUData.subset()

        Returns:
            (pandas.DataFrame)
                Index
                    reset index
                Columns
                    - Date (pd.Timestamp): Observation date
                    - Confirmed (int): the number of confirmed cases
                    - Infected (int): the number of currently infected cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases (> 0)
                    - Susceptible (int): the number of susceptible cases, if calculated

        Note:
            If country is None, the name of the model will be used.
            If province is None, '-' will be used.
            If @population is not None, the number of susceptible cases will be calculated.
            Records with Recovered > 0 will be selected.
        """
        country, _ = self._model_to_area(model=model, country=country, province=province)
        kwargs["population"] = self._population
        kwargs = find_args([super().subset], **kwargs)
        return super().subset(country=country, province=province, **kwargs)

    def subset_complement(self, **kwargs):
        """
        This is the same as ExampleData.subset().
        Complement will not be done.
        """
        return (self.subset(**kwargs), False)

    def records(self, **kwargs):
        """
        This is the same as ExampleData.subset().
        Complement will not be done.
        """
        return (self.subset(**kwargs), False)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from datetime import datetime, timedelta
from functools import partial
import math
import optuna
import numpy as np
import pandas as pd
from scipy.integrate import solve_ivp
from covsirphy.util.error import UnExpectedNoneError, NotNoneError
from covsirphy.util.stopwatch import StopWatch
from covsirphy.util.evaluator import Evaluator
from covsirphy.util.validator import Validator
from covsirphy.util.term import Term


class ODEModel(Term):
    """Basic class of ordinary differential equation (ODE) model.

    Args:
        date_range (tuple of (str, str)): start date and end date of simulation
        tau (int): tau value [min]
        initial_dict (dict of {str: int}): initial values
        param_dict (dict of {str: float}): non-dimensional parameter values
    """
    _logS = "log10(S)"
    _r = "R"
    # Name of ODE model
    _NAME = "ODE Model"
    # Variables
    _VARIABLES = []
    # Non-dimensional parameters
    _PARAMETERS = []
    # Dimensional parameters
    _DAY_PARAMETERS = []
    # Variables that increases monotonically
    _VARS_INCREASE = []
    # Sample data
    _SAMPLE_DICT = {
        "initial_dict": dict.fromkeys(_VARIABLES),
        "param_dict": dict.fromkeys(_PARAMETERS)
    }

    def __init__(self, date_range, tau, initial_dict, param_dict):
        start_date, end_date = Validator(date_range, "date_range", accept_none=False).sequence(length=2)
        self._start = Validator(start_date, name="the first value of @date_range", accept_none=False).date()
        self._end = Validator(
            end_date, name="the second date of @date_range", accept_none=False).date(value_range=(self._start, None))
        self._tau = Validator(tau, "tau", accept_none=False).tau()
        self._initial_dict = Validator(initial_dict, "initial_dict", accept_none=False).dict(
            required_keys=self._VARIABLES, errors="raise")
        self._param_dict = Validator(param_dict, "param_dict", accept_none=False).dict(
            required_keys=self._PARAMETERS, errors="raise")
        # Total population
        self._population = sum(initial_dict.values())
        # Information regarding ODE parameter estimation
        self._estimation_dict = None

    def __str__(self):
        return self._NAME

    def __repr__(self):
        _dict = self.settings()
        return f"{type(self).__name__}({', '.join([f'{k}={v}' for k, v in _dict.items()])})"

    def __eq__(self, other):
        return repr(self) == repr(other)

    @classmethod
    def name(cls):
        """Return name of ODE model.
        """
        return cls._NAME

    @classmethod
    def definitions(cls):
        """Return definitions of ODE model.

        Returns:
            dict of {str: object}:
                - "name" (str): ODE model name
                - "variables" (list of [str]): variable names
                - "parameters" (list of [str]): non-dimensional parameter names
                - "dimensional_parameters" (list of [str]): dimensional parameter names
        """
        return {
            "name": cls._NAME,
            "variables": cls._VARIABLES,
            "parameters": cls._PARAMETERS,
            "dimensional_parameters": cls._DAY_PARAMETERS,
        }

    def settings(self, with_estimation=False):
        """Return settings.

        Args:
            with_estimation (bool): whether includes information regarding ODE parameter estimation or not

        Returns:
            dict of {str: object}:
                - date_range (tuple of (str, str)): start date and end date of simulation
                - tau (int): tau value [min]
                - initial_dict (dict of {str: int}): initial values
                - param_dict (dict of {str: float}): non-dimensional parameter values
                - estimation_dict (dict of {str: str or int}: information regarding ODE parameter estimation, when @with_estimation is True
                    - method (str): method of estimation, "with_quantile" or "with_optimization" or "not_performed"
                    - {metrics} (int): score of hyperparameter optimization, if available
                    - Trials (int) : the number of trials of hyperparameter optimization, if available
                    - Runtime (str): runtime of hyperparameter optimization, if available
                    - keyword arguments set with covsirphy.ODEModel.with_optimization(), if available
                    - keyword arguments set with covsirphy.ODEModel.with_quantile(), if available
        """
        _dict = {
            "date_range": (self._start.strftime(self.DATE_FORMAT), self._end.strftime(self.DATE_FORMAT)),
            "tau": self._tau,
            "initial_dict": self._initial_dict,
            "param_dict": self._param_dict,
        }
        if with_estimation:
            _dict["estimation_dict"] = self._estimation_dict or {"method": "not_performed"}
        return _dict

    @classmethod
    def from_sample(cls, date_range=None, tau=1440):
        """Initialize model with sample data.

        Args:
            date_range (tuple(str or None, str or None) or None): start date and end date of simulation
            tau (int): tau value [min]

        Returns:
            covsirphy.ODEModel: initialized model

        Note:
            When @date_range or the first value of @date_range is None, today when executed will be set as start date.

        Note:
            When @date_range or the second value of @date_range is None, 180 days after start date will be used as end date.
        """
        start_date, end_date = Validator(date_range, "date_range").sequence(default=(None, None), length=2)
        start = Validator(start_date, name="the first value of @date_range").date(default=datetime.now())
        end = Validator(
            end_date, name="the second date of @date_range").date(value_range=(start, None), default=start + timedelta(days=180))
        return cls(date_range=(start, end), tau=tau, **cls._SAMPLE_DICT)

    def _discretize(self, t, X):
        """Discretize the ODE.

        Args:
            t (int): discrete time-steps
            X (numpy.array): the current values of the model

        Returns:
            numpy.array: the next values of the model
        """
        raise NotImplementedError

    def solve(self):
        """Solve an initial value problem.

        Return:
            pandas.DataFrame: analytical solution
                Index
                    Date (pandas.Timestamp): dates from start date to end date
                Columns
                    (pandas.Int64): model-specific dimensional variables of the model
        """
        step_n = math.ceil((self._end - self._start) / timedelta(minutes=self._tau))
        sol = solve_ivp(
            fun=self._discretize,
            t_span=[0, step_n],
            y0=np.array([self._initial_dict[variable] for variable in self._VARIABLES]),
            t_eval=np.arange(0, step_n + 1, 1),
            dense_output=False
        )
        df = pd.DataFrame(data=sol["y"].T.copy(), columns=self._VARIABLES)
        df = self._non_dim_to_date(data=df, tau=self._tau, start_date=self._start)
        return df.round().convert_dtypes()

    @staticmethod
    def _date_to_non_dim(series, tau):
        """Convert date information (TIME) to time(x) = (TIME(x) - TIME(0)) / tau

        Args:
            series (pandas.DatetimeIndex): date information
            tau (int or None): tau value [min]

        Returns:
            pandas.DatetimeIndex: as-is @series when tau is None else converted time information without series name
        """
        Validator(series, "index of data").instance(pd.DatetimeIndex)
        if tau is None:
            return series
        Validator(tau, "tau", accept_none=False).tau()
        converted = (series - series.min()) / np.timedelta64(tau, "m")
        return converted.rename(None).astype("Int64")

    @classmethod
    def _non_dim_to_date(cls, data, tau, start_date):
        """Convert non-dimensional date information (time) to TIME(x) = TIME(0) + tau * time(x) and resample with dates.

        Args:
            data (pandas.DataFrame):
                Index
                    non-dimensional date information pandas.DatetimeIndex
                Columns
                    any columns
            tau (int or None): tau value [min]
            start_date (str or pandas.Timestamp or None): start date of records ie. TIME(0)

        Raises:
            NotNoneError: @tau is None, but start_date is not None
            UnExpectedNoneError: @tau is not None, but start_date is None

        Returns:
            pandas.DataFrame:
                Index
                    Date (pandas.DatetimeIndex) or as-is @data (when @tau is None)
                Columns
                    any columns of @data
            pandas.DatetimeIndex: as-is @series when tau is None else converted time information without series name

        Note:
            The first values on date will be selected when resampling.
        """
        df = Validator(data, "data").dataframe()
        if (tau is None and start_date is None) or isinstance(df.index, pd.DatetimeIndex):
            return data
        if tau is None:
            raise NotNoneError("start_date", start_date, details="None is required because tau is None")
        if start_date is None:
            raise UnExpectedNoneError("start_date", details="Not None value is required because tau is not None")
        Validator(tau, "tau", accept_none=False).tau()
        start = Validator(start_date, "start_date", accept_none=False).date()
        df[cls.DATE] = (start + pd.Series(df.index * tau).apply(lambda x: timedelta(minutes=x))).tolist()
        return df.set_index(cls.DATE).resample("D").first()

    @classmethod
    def transform(cls, data, tau=None):
        """Transform a dataframe, converting Susceptible/Infected/Fatal/Recovered to model-specific variables.

        Args:
            data (pandas.DataFrame):
                Index
                    reset index or pandas.DatetimeIndex (when tau is not None)
                Columns
                    - Susceptible (int): the number of susceptible cases
                    - Infected (int): the number of currently infected cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
            tau (int or None): tau value [min]


        Returns:
            pandas.DataFrame:
                Index
                    as the same as index of @data when @tau is None else converted to time(x) = (TIME(x) - TIME(0)) / tau
                Columns
                    model-specific variables

        Note:
            This method must be defined by child classes.
        """
        raise NotImplementedError

    @classmethod
    def inverse_transform(cls, data, tau=None, start_date=None):
        """Transform a dataframe, converting model-specific variables to Susceptible/Infected/Fatal/Recovered.

        Args:
            data (pandas.DataFrame):
                Index
                    any index
                Columns
                    model-specific variables
            tau (int or None): tau value [min]
            start_date (str or pandas.Timestamp or None): start date of records ie. TIME(0)

        Returns:
            pandas.DataFrame:
                Index
                    Date (pandas.DatetimeIndex) or as-is @data (when either @tau or @start_date are None)
                Columns
                    - Susceptible (int): the number of susceptible cases
                    - Infected (int): the number of currently infected cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases

        Note:
            This method must be defined by child classes.
        """
        raise NotImplementedError

    def r0(self):
        """Calculate basic reproduction number.

        Returns:
            float: reproduction number of the ODE model and parameters

        Note:
            This method must be defined by child classes.
        """
        raise NotImplementedError

    def dimensional_parameters(self):
        """Calculate dimensional parameter values.

        Returns:
            dict of {str: int or float}: dictionary of dimensional parameter values

        Note:
            This method must be defined by child classes.
        """
        raise NotImplementedError

    @classmethod
    def from_data(cls, data, param_dict, tau=1440, digits=None):
        """Initialize model with data and ODE parameter values.

        Args:
            data (pandas.DataFrame):
                Index
                    reset index
                Columns
                    - Date (pd.Timestamp): Observation date
                    - Susceptible (int): the number of susceptible cases
                    - Infected (int): the number of currently infected cases
                    - Recovered (int): the number of recovered cases
                    - Fatal (int): the number of fatal cases
            param_dict (dict of {str: float}): non-dimensional parameter values
            tau (int): tau value [min]
            digits (int or None): effective digits of ODE parameter values or None (skip rounding)

        Returns:
            covsirphy.ODEModel: initialized model
        """
        Validator(data, "data", accept_none=False).dataframe(columns=[cls.DATE, *cls._SIRF], empty_ok=False)
        Validator(tau, "tau", accept_none=False).tau()
        Validator(param_dict, "param_dict", accept_none=False).dict(required_keys=cls._PARAMETERS)
        start, end = data[cls.DATE].min(), data[cls.DATE].max()
        trans_df = cls.transform(data=data.set_index(cls.DATE), tau=tau)
        initial_dict = trans_df.iloc[0].to_dict()
        return cls(
            date_range=(start, end), tau=tau, initial_dict=initial_dict,
            param_dict=param_dict if digits is None else {k: Validator(v, k).float(
                value_range=(0, 1), digits=digits) for k, v in param_dict.items()}
        )

    @classmethod
    def from_data_with_quantile(cls, data, tau=1440, q=0.5, digits=None):
        """Initialize model with data, estimating ODE parameters with quantiles.

        Args:
            data (pandas.DataFrame):
                Index
                    reset index
                Columns
                    - Date (pd.Timestamp): Observation date
                    - Susceptible (int): the number of susceptible cases
                    - Infected (int): the number of currently infected cases
                    - Recovered (int): the number of recovered cases
                    - Fatal (int): the number of fatal cases
            tau (int): tau value [min]
            q (float): the quantiles to compute, values between (0, 1)
            digits (int or None): effective digits of ODE parameter values or None (skip rounding)

        Returns:
            covsirphy.ODEModel: initialized model
        """
        Validator(data, "data", accept_none=False).dataframe(columns=[cls.DATE, *cls._SIRF], empty_ok=False)
        Validator(tau, "tau", accept_none=False).tau()
        Validator(q, "q", accept_none=False).float(value_range=(0, 1))
        trans_df = cls.transform(data=data.set_index(cls.DATE), tau=tau)
        cls_obj = cls.from_data(data=data, param_dict=cls._param_quantile(data=trans_df, q=q), tau=tau, digits=digits)
        cls_obj._estimation_dict = dict(method="with_quantile", tau=tau, q=q, digits=digits)
        return cls_obj

    @classmethod
    def _param_quantile(cls, data, q=0.5):
        """With combinations (X, dX/dt) for variables, calculate quantile values of ODE parameters.

        Args:
            data (pandas.DataFrame): transformed data with covsirphy.ODEModel.transform(data=data, tau=tau)
            q (float or array-like): the quantile(s) to compute, value(s) between (0, 1)

        Returns:
            dict of {str: float or pandas.Series}: parameter values at the quantile(s)

        Note:
            This method must be defined by child classes.
        """
        raise NotImplementedError

    @classmethod
    def _clip(cls, values, lower, upper):
        """
        Trim values at input threshold.

        Args:
            values (float or array-like): values to trim
            lower (float): minimum threshold
            upper (float): maximum threshold

        Returns:
            float or pandas.Series: clipped array
        """
        return min(max(values, lower), upper) if isinstance(values, float) else pd.Series(values).clip()

    @classmethod
    def from_data_with_optimization(cls, data, tau=1440, metric="RMSLE", digits=None, **kwargs):
        """Initialize model with data, estimating ODE parameters hyperparameter optimization using Optuna.

        Args:
            data (pandas.DataFrame):
                Index
                    reset index
                Columns
                    - Date (pd.Timestamp): Observation date
                    - Susceptible (int): the number of susceptible cases
                    - Infected (int): the number of currently infected cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
            tau (int): tau value [min]
            metric (str): metric to minimize, refer to covsirphy.Evaluator.score()
            digits (int or None): effective digits of ODE parameter values or None (skip rounding)
            **kwargs: keyword arguments of optimization
                - quantiles (tuple(int, int)): quantiles to cut parameter range, like confidence interval, (0.1, 0.9) as default
                - timeout (int): timeout of optimization, 180 as default
                - timeout_iteration (int): timeout of one iteration, 1 as default
                - tail_n (int): the number of iterations to decide whether score did not change for the last iterations, 4 as default
                - allowance (tuple(float, float)): the allowance of the max predicted values, (0.99, 1.01) as default
                - pruner (str): kind of pruner (hyperband, median, threshold or percentile), "threshold" as default
                - upper (float): works for "threshold" pruner, intermediate score is larger than this value, it prunes, 0.5 as default
                - percentile (float): works for "Percentile" pruner, the best intermediate value is in the bottom percentile among trials, it prunes, 50 as default
                - constant_liar (bool): whether use constant liar to reduce search effort or not, False as default

        Returns:
            covsirphy.ODEModel: initialized model
        """
        kwargs_default = {
            "quantiles": (0.1, 0.9),
            "timeout": 180,
            "timeout_iteration": 1,
            "tail_n": 4,
            "allowance": (0.99, 1.01),
            "pruner": "threshold",
            "upper": 0.5,
            "percentile": 50,
            "constant_liar": False,
        }
        kwargs_dict = Validator(kwargs, "kwargs").dict(default=kwargs_default)
        Validator(data, "data", accept_none=False).dataframe(columns=[cls.DATE, *cls._SIRF], empty_ok=False)
        Validator(tau, "tau", accept_none=False).tau()
        Validator([metric], "metric").sequence(candidates=Evaluator.metrics())
        trans_df = cls.transform(data=data.set_index(cls.DATE), tau=tau)
        param_dict, score, n_trials, runtime = cls._estimate_params(
            data=trans_df, tau=tau, metric=metric, **kwargs_dict)
        cls_obj = cls.from_data(data=data, param_dict=param_dict, tau=tau, digits=digits)
        cls_obj._estimation_dict = {
            "method": "with_optimization", metric: score, cls.TRIALS: n_trials, cls.RUNTIME: runtime,
            cls.TAU: tau, "digits": digits, **kwargs_dict, "data": data,
        }
        return cls_obj

    @classmethod
    def _estimate_params(cls, data, tau, metric, **kwargs):
        """Estimate ODE parameter values with hyperparameter optimization.

        Args:
            data (pandas.DataFrame): transformed data with covsirphy.ODEModel.transform(data=data, tau=tau)
            tau (int): tau value [min]
            metric (str): metric to minimize, refer to covsirphy.Evaluator.score()
            **kwargs: keyword arguments of optimization and must includes
                - quantiles (tuple(int, int)): quantiles to cut parameter range, like confidence interval
                - timeout (int): timeout of optimization
                - timeout_iteration (int): timeout of one iteration
                - tail_n (int): the number of iterations to decide whether score did not change for the last iterations
                - allowance (tuple(float, float)): the allowance of the max predicted values
                - pruner (str): kind of pruner (hyperband, median, threshold or percentile)
                - upper (float): works for "threshold" pruner, intermediate score is larger than this value, it prunes
                - percentile (float): works for "Percentile" pruner, the best intermediate value is in the bottom percentile among trials, it prunes
                - constant_liar (bool): whether use constant liar to reduce search effort or not

        Returns:
            tuple of (dict of {str: float}, str, int):
                - dict of {str: float}: dictionary of parameter values
                - float: score with metrics
                - str: runtime of hyperparameter optimization
                - int: the number of trials of hyperparameter optimization
        """
        optuna.logging.set_verbosity(optuna.logging.WARNING)
        # Create study of optimization
        pruner_dict = {
            "hyperband": optuna.pruners.HyperbandPruner,
            "median": optuna.pruners.MedianPruner,
            "threshold": optuna.pruners.ThresholdPruner,
            "percentile": optuna.pruners.PercentilePruner,
        }
        pruner_name = Validator([kwargs["pruner"]], "pruner", accept_none=False).sequence(
            candidates=pruner_dict.keys(), length=1)[0]
        pruner_class = pruner_dict[pruner_name]
        pruner = pruner_class(**Validator(kwargs, "keyword arguments").kwargs(pruner_class))
        sampler = optuna.samplers.TPESampler(
            **Validator(kwargs, "keyword arguments").kwargs(optuna.samplers.TPESampler))
        study = optuna.create_study(direction="minimize", sampler=sampler, pruner=pruner)
        # Create objective function
        value_range_dict = cls._param_quantile(data, q=kwargs["quantiles"])
        objective_func = partial(
            cls._optuna_objective, data=data, tau=tau, value_range_dict=value_range_dict, metric=metric)
        allowance = Validator(kwargs["allowance"], "allowance", accept_none=False).sequence(length=2)
        # Iteration of optimization
        iter_n = math.ceil(kwargs["timeout"] / kwargs["timeout_iteration"])
        stopwatch = StopWatch()
        scores = []
        param_dict = {}
        tail_n = Validator(kwargs["tail_n"], "tail_n").int(value_range=(1, None))
        for _ in range(iter_n):
            # Run iteration
            study.optimize(objective_func, n_jobs=1, timeout=kwargs["timeout_iteration"])
            param_dict = study.best_params.copy()
            # If score did not change in the last iterations, stop running
            scores.append(cls._optuna_score(param_dict, data, tau, metric))
            if len(scores) >= tail_n and len(set(scores[-tail_n:])) == 1:
                break
            # Check max values are in the allowance
            if cls._optuna_is_in_allowance(param_dict, data, tau, allowance):
                break
        return param_dict, scores[-1], len(study.trials), stopwatch.stop_show()

    @classmethod
    def _optuna_objective(cls, trial, data, tau, value_range_dict, metric):
        """Objective function to minimize (evaluation score of the difference of actual data and solved data) by Optuna.

        Args:
            trial (optuna.trial): a trial of the study
            data (pandas.DataFrame): transformed data with covsirphy.ODEModel.transform(data=data, tau=tau)
            tau (int): tau value [min]
            value_range_dict (dict of {str: pandas.Series}): dictionary of value range of ODE parameters
            metric (str): metric to minimize, refer to covsirphy.Evaluator.score()

        Returns:
            float: score to minimize
        """
        param_dict = {}
        for (k, v) in value_range_dict.items():
            try:
                param_dict[k] = trial.suggest_float(k, *v)
            except OverflowError:
                param_dict[k] = trial.suggest_float(k, 0, 1)
        return cls._optuna_score(param_dict, data, tau, metric)

    @classmethod
    def _optuna_score(cls, param_dict, data, tau, metric):
        """Score function to minimize (i.e. evaluation score of the difference of actual data and solved data).

        Args:
            param_dict (dict of {str: float}): non-dimensional parameter values
            data (pandas.DataFrame): transformed data with covsirphy.ODEModel.transform(data=data, tau=tau)
            tau (int): tau value [min]
            metric (str): metric to minimize, refer to covsirphy.Evaluator.score()

        Returns:
            float: score to minimize or positive infinity (when negative values are included in simulated values)
        """
        df = cls._non_dim_to_date(data=data, tau=tau, start_date="01Jan2022")
        model = cls.from_data(data=cls.inverse_transform(df).reset_index(), param_dict=param_dict, tau=tau, digits=None)
        sim_df = model.solve()
        evaluator = Evaluator(df, sim_df, how="inner", on=None)
        return evaluator.score(metric=metric)

    @classmethod
    def _optuna_is_in_allowance(cls, param_dict, data, tau, allowance):
        """
        Return whether all max values of estimated values are in allowance or not.

        Args:
            param_dict (dict of {str: float}): non-dimensional parameter values
            data (pandas.DataFrame): transformed data with covsirphy.ODEModel.transform(data=data, tau=tau)
            tau (int): tau value [min]
            allowance (tuple(float, float)): the allowance of the predicted value

        Returns:
            bool: True when all max values of predicted values are in allowance
        """
        df = cls.inverse_transform(data=data, tau=tau, start_date="01Jan2022").reset_index()
        max_dict = {v: data[v].max() for v in cls._VARIABLES}
        model = cls.from_data(data=df, param_dict=param_dict, tau=tau, digits=None)
        sim_df = model.solve()
        sim_max_dict = {v: sim_df[v].max() for v in cls._VARIABLES}
        # Check all max values are in allowance
        allowance0, allowance1 = allowance
        ok_list = [
            a * allowance0 <= p <= a * allowance1 for (a, p) in zip(max_dict.values(), sim_max_dict.values())]
        return all(ok_list)

    @classmethod
    def sr(cls, data):
        """Return log10(S) and R of model-specific variables for S-R trend analysis.

        Args:
            data (pandas.DataFrame):
                Index
                    Date (pd.Timestamp): Observation date
                Columns
                    Susceptible (int): the number of susceptible cases
                    Infected (int): the number of currently infected cases
                    Recovered (int): the number of recovered cases
                    Fatal (int): the number of fatal cases

        Returns:
            pandas.DataFrame:
                Index
                    Date (pandas.Timestamp): date
                Columns
                    log10(S) (np.float64): common logarithm of S of the ODE model
                    R (np.int64): R of the model
        """
        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import contextlib
from pathlib import Path
import pandas as pd
from covsirphy.util.validator import Validator
from covsirphy.util.term import Term


class _RemoteDatabase(Term):
    """
    Base class to access remote databases.

    Args:
        filename (str): CSV filename to save records
        iso3 (str or None): ISO3 code of the country which must be included in the dataset or None (all available countries)
    """
    # Citation
    CITATION = ""
    # Column names and data types
    # {"name in database": "name defined in Term class"}
    COL_DICT = {}

    def __init__(self, filename, iso3):
        # Filepath to save files
        try:
            self.filepath = Path(filename)
        except TypeError as e:
            raise TypeError(f"@filename should be a path-like object, but {filename} was applied.") from e

        self.filepath.parent.mkdir(exist_ok=True, parents=True)
        # Country
        self._iso3 = None if iso3 is None else str(iso3)
        # List of column names (defined in Term class)
        self.saved_cols = list(self.COL_DICT.values())

    def to_dataframe(self, force, verbose):
        """
        Load the dataset and return it as a dataframe.

        Args:
            force (bool): whether download the dataset from server or not when we have saved dataset
            verbose (int): level of verbosity

        Returns:
            pandas.DataFrame
                Index
                    reset index
                Columns
                    defined by the first values of self.COL_DICT.values()
        """
        # Read local file if available and usable
        if not force and self.filepath.exists():
            with contextlib.suppress(ValueError):
                df = self._ensure_dataframe(self.read(), columns=self.saved_cols)
                if self._iso3 is None and df[self.ISO3].nunique() > 1:
                    return df
                if self._iso3 is not None:
                    Validator(df, "raw data").dataframe()
                    if self._value_included(df, value_dict={self.ISO3: self._iso3}):
                        return df
        # Download dataset from server
        df = self.download(verbose=verbose)
        df = df.rename(columns=self.COL_DICT)
        df = self._ensure_dataframe(df, columns=self.saved_cols)
        df.to_csv(self.filepath, index=False)
        return df

    def _value_included(self, target, value_dict):
        """
        Return whether all expected values are included in the columns or not.

        Args:
            target (pandas.DataFrame): the dataframe to ensure
            value_dict (dict[str, object]): dictionary of values which must be included in the column

        Returns:
            bool: whether all values (specified in @value_dict) are included or not
        """
        return all(value in target[col].unique() for (col, value) in value_dict.items()) if set(value_dict.keys()).issubset(target.columns) else False

    @staticmethod
    def _ensure_dataframe(target, name="df", time_index=False, columns=None, empty_ok=True):
        """
        Ensure the dataframe has the columns.

        Args:
            target (pandas.DataFrame): the dataframe to ensure
            name (str): argument name of the dataframe
            time_index (bool): if True, the dataframe must has DatetimeIndex
            columns (list[str] or None): the columns the dataframe must have
            empty_ok (bool): whether give permission to empty dataframe or not

        Returns:
            pandas.DataFrame:
                Index
                    as-is
                Columns:
                    columns specified with @columns or all columns of @target (when @columns is None)
        """
        if not isinstance(target, pd.DataFrame):
            raise TypeError(f"@{name} must be a instance of (pandas.DataFrame).")
        df = target.copy()
        if time_index and (not isinstance(df.index, pd.DatetimeIndex)):
            raise TypeError(f"Index of @{name} must be <pd.DatetimeIndex>.")
        if not empty_ok and target.empty:
            raise ValueError(f"@{name} must not be a empty dataframe.")
        if columns is None:
            return df
        if not set(columns).issubset(df.columns):
            cols_str = ", ".join(col for col in columns if col not in df.columns)
            included = ", ".join(df.columns.tolist())
            s1 = f"Expected columns were not included in {name} with {included}."
            raise KeyError(f"{s1} {cols_str} must be included.")
        return df.loc[:, columns]

    def read(self):
        """
        Load the dataset with a local file.

        Returns:
            pandas.DataFrame
                Index
                    reset index
                Columns
                    defined by .COL_DICT.values()
        """
        kwargs = {"low_memory": False, "header": 0, "usecols": self.saved_cols}
        return pd.read_csv(self.filepath, **kwargs)

    def download(self, verbose):
        """
        Download the dataset from the server and set the list of primary sources.

        Args:
            verbose (int): level of verbosity

        Returns:
            pandas.DataFrame
                Index
                    reset index
                Columns
                    defined by .COL_DICT.values()
        """
        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pandas as pd
from covsirphy.util.error import SubsetNotFoundError, deprecate
from covsirphy.cleaning.cbase import CleaningBase


class OxCGRTData(CleaningBase):
    """
    Data cleaning of OxCGRT dataset.

    Args:
        filename (str or None): CSV filename of the dataset
        data (pandas.DataFrame or None):
            Index
                reset index
            Columns
                - Date: Observation date
                - ISO3: ISO 3166-1 alpha-3, like JPN
                - Country: country/region name
                - variables defined by @variables
        citation (str or None): citation or None (empty)
        variables (list[str] or None): variables to parse or None (use default variables listed as follows)
            - School_closing
            - Workplace_closing
            - Cancel_events
            - Gatherings_restrictions
            - Transport_closing
            - Stay_home_restrictions
            - Internal_movement_restrictions
            - International_movement_restrictions
            - Information_campaigns
            - Testing_policy
            - Contact_tracing
            - Stringency_index

    Note:
        Either @filename (high priority) or @data must be specified.

    Note:
        The default policy indices (Overall etc.) are from README.md and documentation/index_methodology.md in
        https://github.com/OxCGRT/covid-policy-tracker/
    """
    OXCGRT_VARS = [
        "School_closing",
        "Workplace_closing",
        "Cancel_events",
        "Gatherings_restrictions",
        "Transport_closing",
        "Stay_home_restrictions",
        "Internal_movement_restrictions",
        "International_movement_restrictions",
        "Information_campaigns",
        "Testing_policy",
        "Contact_tracing",
        "Stringency_index"
    ]
    # Indicators except for Stringency index
    OXCGRT_VARS_INDICATORS = [v for v in OXCGRT_VARS if v != "Stringency_index"]

    @deprecate(old="OxCGRTData", new="DataEngineer", version="2.24.0-xi")
    def __init__(self, filename=None, data=None, citation=None, variables=None):
        self._variables = variables or self.OXCGRT_VARS[:]
        super().__init__(filename=filename, data=data, citation=citation, variables=self._variables)

    def _cleaning(self):
        """
        Perform data cleaning of the raw data.

        Returns:
            pandas.DataFrame
                Index
                    reset index
                Columns
                    - Date (pandas.Timestamp): Observation date
                    - ISO3 (str): ISO 3166-1 alpha-3, like JPN
                    - Country (pandas.Category): country/region name
                    - Province (pandas.Category): province/prefecture/state name
                    - variables defined by OxCGRTData(variables)
        """
        df = self._raw.copy()
        # Prepare data for Greenland
        grl_df = df.loc[df[self.COUNTRY] == "Denmark"].copy()
        grl_df.loc[:, [self.ISO3, self.COUNTRY]] = ["GRL", "Greenland"]
        df = pd.concat([df, grl_df], sort=True, ignore_index=True)
        # Confirm the expected columns are in raw data
        self._ensure_dataframe(df, name="the raw data", columns=self._raw_cols)
        # Read date records
        df[self.DATE] = pd.to_datetime(df[self.DATE])
        # Confirm float type
        for col in self._variables:
            df[col] = pd.to_numeric(df[col], errors="coerce").ffill()
        # Update data types to reduce memory
        cat_cols = [self.ISO3, self.COUNTRY, self.PROVINCE]
        df[cat_cols] = df[cat_cols].astype("category")
        return df.loc[:, self._raw_cols]

    def subset(self, country, province=None, **kwargs):
        """
        Create a subset for a country.

        Args:
            country (str): country name or ISO 3166-1 alpha-3, like JPN
            province (str): province name
            kwargs: the other arguments will be ignored at the latest version

        Raises:
            covsirphy.SubsetNotFoundError: no records were found

        Returns:
            pandas.DataFrame
                Index
                    reset index
                Columns
                    - Date (pandas.Timestamp): Observation date
                    - variables defined by OxCGRTData(variables)
        """
        country_arg = country
        country = self.ensure_country_name(country)
        try:
            df = super().subset(country=country, province=province)
        except SubsetNotFoundError:
            raise SubsetNotFoundError(
                country=country_arg, country_alias=country, province=province) from None
        df = df.groupby(self.DATE).last().reset_index()
        return df.loc[:, self._subset_cols]

    def total(self):
        """
        This method is not defined for OxCGRTData class.
        """
        raise NotImplementedError

    def map(self, country=None, variable="Stringency_index", date=None, **kwargs):
        """
        Create global colored map to show the values.

        Args:
            country (None): always None
            variable (str): variable name to show
            date (str or None): date of the records or None (the last value)
            kwargs: arguments of ColoredMap() and ColoredMap.plot()
        """
        # Date
        date_str = date or self.cleaned()[self.DATE].max().strftime(self.DATE_FORMAT)
        country_str = country or "Global"
        title = f"{country_str}: {variable.lower().replace('_', ' ')} on {date_str}"
        # Global map
        if country is None:
            return self._colored_map_global(
                variable=variable, title=title, date=date, logscale=False, **kwargs)
        # Country-specific map
        return self._colored_map_country(
            country=country, variable=variable, title=title, date=date, logscale=False, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from covsirphy.util.validator import Validator
from covsirphy.util.term import Term


class Alias(Term):
    """Remember and parse aliases, just like defaultdict.

    Args:
        target_class (object): class of targets or None (all objects)
    """

    def __init__(self, target_class=None):
        self._dict = {}
        self._target_class = target_class or object

    @classmethod
    def for_variables(cls):
        """Initialize covsirphy.Alias with preset of variable aliases.
        """
        class_obj = cls(target_class=list)
        _dict = {
            "N": [cls.N], "S": [cls.S], "T": [cls.TESTS], "C": [cls.C], "I": [cls.CI], "F": [cls.F], "R": [cls.R],
            "CFR": [cls.C, cls.F, cls.R],
            "CIRF": [cls.C, cls.CI, cls.R, cls.F],
            "SIRF": [cls.S, cls.CI, cls.R, cls.F],
            "CR": [cls.C, cls.R],
        }
        [class_obj.update(name, target) for name, target in _dict.items()]
        return class_obj

    def update(self, name, target):
        """Update target of the alias.

        Args:
            name (str): alias name
            targets (object): target to link with the name

        Return:
            covsirphy.Alias: self
        """
        Validator(name, "name", accept_none=False).instance(str)
        self._dict[name] = Validator(target, "target").instance(expected=self._target_class)
        return self

    def find(self, name, default=None):
        """Find the target of the alias.

        Args:
            name (str): alias name
            default (object): default value when not found

        Returns:
            object: target or default value
        """
        try:
            return self._dict.get(name, default)
        except TypeError:
            return default

    def all(self):
        """List up all targets of aliases.

        Returns:
            dict of {str: object}
        """
        return self._dict

    def delete(self, name):
        """Delete alias.

        Args:
            name (str): alias name

        Raises:
            KeyError: the alias has not been registered as an alias

        Return:
            covsirphy.Alias: self
        """
        try:
            del self._dict[name]
        except KeyError:
            raise KeyError(f"{name} has not been registered as an alias.") from None
        return self



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from datetime import timedelta
from functools import partial
from multiprocessing import cpu_count, Pool
import numpy as np
import pandas as pd
from p_tqdm import p_umap
from covsirphy.util.error import EmptyError, NotEnoughDataError, UnExpectedNoneError, UnExpectedValueRangeError
from covsirphy.util.evaluator import Evaluator
from covsirphy.util.stopwatch import StopWatch
from covsirphy.util.validator import Validator
from covsirphy.util.term import Term
from covsirphy.visualization.compare_plot import compare_plot
from covsirphy.dynamics.ode import ODEModel
from covsirphy.dynamics._trend import _TrendAnalyzer
from covsirphy.dynamics._simulator import _Simulator


class Dynamics(Term):
    """Class to hand phase-dependent SIR-derived ODE models.

    Args:
        model (covsirphy.ODEModel): definition of ODE model
        date_range (tuple of (str, str)): start date and end date of dynamics to analyze
        tau (int or None): tau value [min] or None (set later with data)
        name (str or None): name of dynamics to show in figures (e.g. "baseline") or None (un-set)
    """

    def __init__(self, model, date_range, tau=None, name=None):
        self._model = Validator(model, "model", accept_none=False).subclass(ODEModel)
        first_date, last_date = Validator(date_range, "date_range", accept_none=False).sequence(length=2)
        self._first = Validator(first_date, name="the first value of @date_range", accept_none=False).date()
        self._last = Validator(
            last_date, name="the second date of @date_range", accept_none=False).date(value_range=(self._first, None))
        self._tau = Validator(tau, "tau", accept_none=True).tau()
        self._name = None if name is None else Validator(name, "name").instance(str)
        # Index: Date, Columns: S, I, F, R, ODE parameters
        self._parameters = self._model._PARAMETERS[:]
        self._df = pd.DataFrame(
            {self._PH: 0}, index=pd.date_range(start=self._first, end=self._last, freq="D"),
            columns=[self._PH, *self._SIRF, *self._parameters])

    def __len__(self):
        return self._df[self._PH].nunique()

    @property
    def model(self):
        """covsirphy.ODEModel: model class
        """
        return self._model

    @property
    def model_name(self):
        """str: name of ODE model
        """
        return self._model._NAME

    @property
    def tau(self):
        """int or None: tau value [min] or None (un-set)
        """
        return self._tau

    @tau.setter
    def tau(self, value):
        self._tau = Validator(value, "tau", accept_none=True).tau()

    @tau.deleter
    def tau(self):
        self._tau = None

    @property
    def name(self):
        """str: name of dynamics to show in figures (e.g. "baseline") or None (un-set)
        """
        return self._name

    @name.setter
    def name(self, name):
        self._name = Validator(name, "name").instance(str)

    @name.deleter
    def name(self):
        self._name = None

    @classmethod
    def from_sample(cls, model, date_range=None, tau=1440):
        """Initialize model with sample data of one-phase ODE model.

        Args:
            date_range (tuple(str or None, str or None) or None): start date and end date of simulation
            tau (int): tau value [min]

        Returns:
            covsirphy.ODEModel: initialized model

        Note:
            Regarding @date_range, refer to covsirphy.ODEModel.from_sample().
        """
        Validator(model, "model", accept_none=False).subclass(ODEModel)
        model_instance = model.from_sample(date_range=date_range, tau=tau)
        settings_dict = model_instance.settings()
        variable_df = model.inverse_transform(model_instance.solve()).iloc[[0]]
        param_df = pd.DataFrame(settings_dict["param_dict"], index=[pd.to_datetime(settings_dict["date_range"][0])])
        param_df.index.name = cls.DATE
        df = pd.concat([variable_df, param_df], axis=1)
        instance = cls(model=model, date_range=settings_dict["date_range"], tau=tau, name="Sample data")
        instance.register(data=df)
        return instance

    @classmethod
    def from_data(cls, model, data, tau=1440, name=None):
        """Initialize model with data.

        Args:
            data (pandas.DataFrame): new data to overwrite the current information
                Index
                    Date (pandas.Timestamp): Observation dates
                Columns
                    Susceptible (int): the number of susceptible cases
                    Infected (int): the number of currently infected cases
                    Fatal (int): the number of fatal cases
                    Recovered (int): the number of recovered cases
                    (numpy.float64): ODE parameter values defined with the ODE model (optional)
            tau (int or None): tau value [min] or None (un-set)
            name (str or None): name of dynamics to show in figures (e.g. "baseline") or None (un-set)

        Returns:
            covsirphy.ODEModel: initialized model

        Note:
            Regarding @date_range, refer to covsirphy.ODEModel.from_sample().
        """
        Validator(model, "model", accept_none=False).subclass(ODEModel)
        Validator(data, "data").dataframe(time_index=True)
        instance = cls(model=model, date_range=(data.index.min(), data.index.max()), tau=tau, name=name)
        instance.register(data=data)
        return instance

    def register(self, data=None):
        """Register data to get initial values and ODE parameter values (if available).

        Args:
            data (pandas.DataFrame or None): new data to overwrite the current information or None (no new records)
                Index
                    Date (pandas.Timestamp): Observation dates
                Columns
                    Susceptible (int): the number of susceptible cases
                    Infected (int): the number of currently infected cases
                    Recovered (int): the number of recovered cases
                    Fatal (int): the number of fatal cases
                    (numpy.float64): ODE parameter values defined with the model

        Returns:
            pandas.DataFrame: the current information
                Index
                    Date (pandas.Timestamp): Observation dates
                Columns
                    Susceptible (int): the number of susceptible cases
                    Infected (int): the number of currently infected cases
                    Recovered (int): the number of recovered cases
                    Fatal (int): the number of fatal cases
                    (numpy.float64): ODE parameter values defined with model.PARAMETERS

        Note:
            Change points of ODE parameter values will be recognized as the change points of phases.

        Note:
            NA can used in the newer phases because filled with that of the older phases.
        """
        if data is not None:
            all_df = self._df.copy()
            new_df = Validator(data, "data").dataframe(time_index=True)
            new_df.index = pd.to_datetime(new_df.index).round("D")
            all_df.loc[:] = np.nan
            all_df[self._PH] = 0
            all_df.update(new_df, overwrite=True)
            if all_df.loc[self._first, self._SIRF].isna().any():
                raise EmptyError(
                    f"records on {self._first.strftime(self.DATE_FORMAT)}", details="Records must be registered for simulation")
            if all_df.min().min() < 0:
                raise UnExpectedValueRangeError("minimum value of the data", all_df.min().min(), (0, None))
            all_df.index.name = self.DATE
            self._df = all_df.convert_dtypes()
            # Find change points with parameter values
            param_df = all_df.loc[:, self._parameters].ffill().drop_duplicates().dropna(axis=0)
            if not param_df.empty:
                self._segment(points=param_df.index.tolist(), overwrite=True)
        return self._df.loc[:, [*self._SIRF, *self._parameters]]

    def _segment(self, points, overwrite):
        """Perform time-series segmentation with points.

        Args:
            points (list[str]): dates of change points
            overwrite (bool): whether remove all phases before segmentation or not

        Returns:
            covsirphy.Dynamics: self

        Note:
            @points can include the first date, but not required.

        Note:
            @points must be selected from the first date to three days before the last date specified covsirphy.Dynamics(date_range).
        """
        point_dates = [Validator(point, "a change point", accept_none=False).date() for point in points]
        candidates = pd.date_range(start=self._first, end=self._last - timedelta(days=2), freq="D")
        change_points = Validator(point_dates, "points", accept_none=False).sequence(unique=True, candidates=candidates)
        df = self._df.copy()
        if overwrite:
            df[self._PH] = 0
        for point in change_points:
            df.loc[point:, self._PH] += 1
        self._df = df.convert_dtypes()

    def segment(self, points=None, overwrite=False, **kwargs):
        """Perform time-series segmentation with points manually selected or found with S-R trend analysis.

        Args:
            points (list[str] or None): dates of change points or None (will be found with S-R trend analysis via .detect() method)
            overwrite (bool): whether remove all phases before segmentation or not
            **kwargs: keyword arguments of covsirphy.Dynamics.detect()

        Returns:
            covsirphy.Dynamics: self

        Note:
            @points can include the first date, but not required.

        Note:
            @points must be selected from the first date to three days before the last date specified covsirphy.Dynamics(date_range).
        """
        return self._segment(points=points or self.detect(**kwargs)[0], overwrite=overwrite)

    def detect(self, algo="Binseg-normal", min_size=7, display=True, **kwargs):
        """Perform S-R trend analysis to find change points of log10(S) - R of model-specific variables, not that segmentation requires .segment() method.

        Args:
            algo (str): detection algorithms and models
            min_size (int): minimum value of phase length [days], be equal to or over 3
            display (bool): whether display figure of log10(S) - R plane or not
            **kwargs: keyword arguments of algorithm classes (ruptures.Pelt, .Binseg, BottomUp) except for "model",
                covsirphy.VisualizeBase(), matplotlib.legend.Legend()

        Raises:
            NotEnoughDataError: we have not enough records, the length of the records must be equal to or over min_size * 2

        Returns:
            tuple of (pandas.Timestamp, pandas.DataFrame):
                pandas.Timestamp: date of change points
                pandas.Dataframe:
                    Index
                        R (int): actual R (R of the ODE model) values
                    Columns
                        Actual (float): actual log10(S) (common logarithm of S of the ODE model) values
                        Fitted (float): log10(S) values fitted with y = a * R + b
                        0th (float): log10(S) values fitted with y = a * R + b and 0th phase data
                        1st, 2nd... (float): fitted values of 1st, 2nd phases

        Note:
            Python library `ruptures` will be used for off-line change point detection.
            Refer to documentation of `ruptures` library, https://centre-borelli.github.io/ruptures-docs/
            Candidates of @algo are "Pelt-rbf", "Binseg-rbf", "Binseg-normal", "BottomUp-rbf", "BottomUp-normal".

        Note:
            S-R trend analysis is original to Covsirphy, https://www.kaggle.com/code/lisphilar/covid-19-data-with-sir-model/notebook
            "Phase" means a sequential dates in which the parameters of SIR-derived models are fixed.
            "Change points" means the dates when trend was changed.
            "Change points" is the same as the start dates of phases except for the 0th phase.
        """
        Validator(min_size, "min_size", accept_none=False).int(value_range=(3, None))
        df = self._df.dropna(how="any", subset=self._SIRF)
        if len(df) < min_size * 2:
            raise NotEnoughDataError("the records of the number of cases without NAs", df, required_n=min_size * 2)
        analyzer = _TrendAnalyzer(data=df, model=self._model, min_size=min_size)
        points = analyzer.find_points(algo=algo, **kwargs)
        fit_df = analyzer.fitting(points=points)
        if display:
            analyzer.display(points=points, fit_df=fit_df, name=self._name, **kwargs)
        return points, fit_df

    def summary(self):
        """Summarize phase information.

        Returns:
            pandas.DataFrame:
                Index
                    Phase (str): phase names, 0th, 1st,...
                Columns
                    Start (pandas.Timestamp): start date of the phase
                    End (pandas.Timestamp): end date of the phase
                    Rt (float): phase-dependent reproduction number (if parameters are available)
                    (float): parameter values, including rho (if available)
                    (int or float): dimensional parameters, including 1/beta [days] (if tau and parameters are available)
        """
        df = self._df.reset_index()
        df[self._PH], _ = df[self._PH].factorize()
        first_df = df.groupby(self._PH).first()
        df = first_df.join(df.groupby(self._PH).last(), rsuffix="_last")
        df = df.rename(columns={self.DATE: self.START, f"{self.DATE}_last": self.END})
        df = df.loc[:, [col for col in df.columns if "_last" not in col]]
        df.index = [self.num2str(num) for num in df.index]
        df.index.name = self.PHASE
        # Reproduction number
        df[self.RT] = df[self._parameters].apply(
            lambda x: np.nan if x.isna().any() else self._model.from_data(data=self._df.reset_index(), param_dict=x.to_dict(), tau=self._tau).r0(), axis=1)
        # Day parameters
        if self._tau is not None:
            days_df = df[self._parameters].apply(
                lambda x: np.nan if x.isna().any() else self._model.from_data(
                    data=self._df.reset_index(), param_dict=x.to_dict(), tau=self._tau).dimensional_parameters(),
                axis=1, result_type="expand"
            )
            df = pd.concat([df, days_df], axis=1)
        # Set the order of columns
        fixed_cols = [
            self.START, self.END, self.RT, *self._model._PARAMETERS, *self._model._DAY_PARAMETERS]
        others = [col for col in df.columns if col not in set(fixed_cols) | set(self._SIRF)]
        return df.reindex(columns=[*fixed_cols, *others]).dropna(how="all", axis=1).ffill().convert_dtypes()

    def track(self):
        """Track reproduction number, parameter value and dimensional parameter values.

        Returns:
            pandas.DataFrame
                Index
                    Date (pandas.Timestamp): dates
                Columns
                    Rt (float): phase-dependent reproduction number (if parameters are available)
                    (float): parameter values, including rho (if available)
                    (int or float): dimensional parameters, including 1/beta [days] (if tau and parameters are available)
        """
        df = self.summary()
        df[self.DATE] = df[[self.START, self.END]].apply(
            lambda x: pd.date_range(start=x[0], end=x[1], freq="D"), axis=1)
        return df.explode(self.DATE).set_index(self.DATE).drop([self.START, self.END], axis=1)

    def simulate(self, model_specific=False):
        """Perform simulation with phase-dependent ODE model.

        Args:
            model_specific (bool): whether convert S, I, F, R to model-specific variables or not

        Raises:
            UnExpectedNoneError: tau value is un-set
            NAFoundError: ODE parameter values on the start dates of phases are un-set

        Returns:
            pandas.DataFrame:
                Index
                    Date (pd.Timestamp): dates
                Columns
                    if @model_specific is False:
                    Susceptible (int): the number of susceptible cases
                    Infected (int): the number of currently infected cases
                    Recovered (int): the number of recovered cases
                    Fatal (int): the number of fatal cases
                    if @model_specific is True, variables defined by model.VARIABLES of covsirphy.Dynamics(model)
        """
        if self._tau is None:
            raise UnExpectedNoneError(
                "tau", details="Tau value must be set with covsirphy.Dynamics(tau) or covsirphy.Dynamics.tau or covsirphy.Dynamics.estimate_tau()")
        simulator = _Simulator(model=self._model, data=self._df)
        return simulator.run(tau=self._tau, model_specific=model_specific).set_index(self.DATE)

    def estimate(self, **kwargs):
        """Run covsirphy.Dynamics.estimate_tau() and covsirphy.Dynamics.estimate_params().

        Args:
            **kwargs: keyword arguments of covsirphy.Dynamics.estimate_tau() and covsirphy.Dynamics.estimate_params()

        Returns:
            covsirphy.Dynamics: self
        """
        self.estimate_tau(**Validator(kwargs).kwargs(self.estimate_tau))
        self.estimate_params(**kwargs)
        return self

    def estimate_tau(self, metric="RMSLE", q=0.5, digits=None, n_jobs=None):
        """Set the best tau value for the registered data, estimating ODE parameters with quantiles.

        Args:
            metric (str): metric name for scoring when selecting best tau value
            q (float): the quantiles to compute, values between (0, 1)
            digits (int or None): effective digits of ODE parameter values or None (skip rounding)
            n_jobs (int or None): the number of parallel jobs or None (CPU count)

        Raises:
            NotEnoughDataError: less than three non-NA records are registered

        Returns:
            tuple of (float, pandas.DataFrame):
                float: tau value with best metric score
                pandas.DataFrame: metric scores of tau candidates
                    Index
                        tau (int): candidate of tau values
                    Columns
                        {metric}: score of estimation with metric
        """
        all_df = self._df.dropna(how="any", subset=self._SIRF)
        if len(all_df) < 3:
            raise NotEnoughDataError("registered S/I/F/R data except NAs", all_df, 3)
        score_f = partial(self._score_with_tau, metric=metric, q=q, digits=digits)
        divisors = [i for i in range(1, 1441) if 1440 % i == 0]
        n_jobs_validated = Validator(n_jobs, "n_jobs").int(value_range=(1, cpu_count()), default=cpu_count())
        with Pool(n_jobs_validated) as p:
            scores = p.map(score_f, divisors)
        score_dict = dict(zip(divisors, scores))
        comp_f = {True: min, False: max}[Evaluator.smaller_is_better(metric=metric)]
        self._tau = comp_f(score_dict.items(), key=lambda x: x[1])[0]
        return self._tau, pd.DataFrame.from_dict(score_dict, orient="index", columns=[metric])

    def _score_with_tau(self, tau, metric, q, digits):
        """Return the metric score with tau.

        Args:
            tau (int): tau value [min]
            metric (str): metric name for scoring when selecting best tau value
            q (float): the quantiles to compute, values between (0, 1)
            digits (int or None): effective digits of ODE parameter values or None (skip rounding)

        Returns:
            float: score to minimize
        """
        parameters = self._model._PARAMETERS[:]
        all_df = self._df.dropna(how="any", subset=self._SIRF)
        all_df.loc[:, parameters] = all_df.loc[:, parameters].astype("Float64")
        starts = all_df.reset_index().groupby(self._PH)[self.DATE].first().sort_values()
        ends = all_df.reset_index().groupby(self._PH)[self.DATE].last().sort_values()
        for start, end in zip(starts, ends):
            model_instance = self._model.from_data_with_quantile(
                data=all_df.loc[start: end].reset_index(), tau=tau, q=q, digits=digits)
            all_df.loc[start, parameters] = pd.Series(model_instance.settings()["param_dict"])
        simulator = _Simulator(model=self._model, data=all_df)
        sim_df = simulator.run(tau=tau, model_specific=False).set_index(self.DATE)
        evaluator = Evaluator(all_df[self._SIRF], sim_df[self._SIRF], how="inner")
        return evaluator.score(metric=metric)

    def estimate_params(self, metric="RMSLE", digits=None, n_jobs=None, **kwargs):
        """Set ODE parameter values optimized for the registered data with hyperparameter optimization using Optuna.

        Args:
            metric (str): metric name for scoring when optimizing ODE parameter values of phases
            digits (int or None): effective digits of ODE parameter values or None (skip rounding)
            n_jobs (int or None): the number of parallel jobs or None (CPU count)
            **kwargs: keyword arguments of optimization, refer to covsirphy.ODEModel.from_data_with_optimization()

        Raises:
            UnExpectedNoneError: tau value is un-set
            NotEnoughDataError: less than three non-NA records are registered

        Returns:
            pandas.DataFrame:
                Index
                    Date (pandas.Timestamp): dates
                Columns
                    (numpy.float64): ODE parameter values defined with model.PARAMETERS
                    {metric}: score with the estimated parameter values
                    Trials (int): the number of trials
                    Runtime (str): runtime of optimization, like 0 min 10 sec
        """
        if self._tau is None:
            raise UnExpectedNoneError(
                "tau", details="Tau value must be set with covsirphy.Dynamics(tau) or covsirphy.Dynamics.tau or covsirphy.Dynamics.estimate_tau()")
        all_df = self._df.loc[:, [self._PH, *self._SIRF]].dropna(how="any")
        if len(all_df) < 3:
            raise NotEnoughDataError("registered S/I/F/R data except NAs", all_df, 3)
        n_jobs_validated = Validator(n_jobs, "n_jobs").int(value_range=(1, cpu_count()), default=cpu_count())
        starts = all_df.reset_index().groupby(self._PH)[self.DATE].first().sort_values()
        ends = all_df.reset_index().groupby(self._PH)[self.DATE].last().sort_values()
        est_f = partial(
            self._optimized_params, model=self._model, tau=self._tau, metric=metric, digits=digits, **kwargs)
        phase_dataframes = [all_df[start: end] for start, end in zip(starts, ends)]
        print(f"\n<{self._model._NAME}: parameter estimation>")
        print(f"Running optimization with {n_jobs_validated} CPUs...")
        stopwatch = StopWatch()
        results = p_umap(est_f, phase_dataframes, num_cpus=n_jobs_validated)
        print(f"Completed optimization. Total: {stopwatch.stop_show()}\n")
        est_df = pd.concat(results, sort=True, axis=0)
        est_df = est_df.loc[:, [*self._parameters, metric, self.TRIALS, self.RUNTIME]].ffill().convert_dtypes()
        # Update registered parameter values
        r_df = self.register()
        r_df.update(est_df, overwrite=True)
        self.register(data=r_df)
        return est_df

    def _optimized_params(self, phase_df, model, tau, metric, digits, **kwargs):
        """Return ODE parameter values optimized with the registered data, estimating ODE parameters hyperparameter optimization using Optuna.

        Args:
            metric (str): metric name for scoring when optimizing ODE parameter values of phases
            digits (int or None): effective digits of ODE parameter values or None (skip rounding)
            n_jobs (int or None): the number of parallel jobs or None (CPU count)
            **kwargs: keyword arguments of optimization, refer to covsirphy.ODEModel.from_data_with_optimization()

        Raises:
            UnExpectedNoneError: tau value is un-set

        Returns:
            pandas.DataFrame:
                Index
                    Date (pandas.Timestamp): dates
                Columns
                    (numpy.float64): ODE parameter values defined with model.PARAMETERS
                    {metric}: score with the estimated parameter values
                    Trials (int): the number of trials
                    Runtime (str): runtime of optimization, like 0 min 10 sec
        """
        df = phase_df.copy()
        # ODE parameter optimization
        model_instance = model.from_data_with_optimization(
            data=df.reset_index(), tau=tau, metric=metric, digits=digits, **kwargs)
        df.loc[df.index[0], model._PARAMETERS] = pd.Series(model_instance.settings()["param_dict"])
        # Get information regarding optimization
        est_dict = model_instance.settings(with_estimation=True)["estimation_dict"]
        est_dict = {k: v for k, v in est_dict.items() if k in {metric, self.TRIALS, self.RUNTIME}}
        df.loc[df.index[0], list(est_dict.keys())] = pd.Series(est_dict)
        return df

    def parse_phases(self, phases=None):
        """Return minimum date and maximum date of the phases.

        Args:
            phases (list of [str], or None): phases (0th, 1st, 2nd,... last) or None (all phases)

        Returns:
            tuple of (pandas.Timestamp): minimum date and maximum date of the phases

        Note:
            "last" can be used to specify the last phase.
        """
        if phases is None:
            return self._first, self._last
        all_df = self._df.copy()
        all_df[self._PH], _ = all_df[self._PH].factorize()
        phase_numbers = [all_df[self._PH].max() if ph == "last" else self.str2num(ph) for ph in phases]
        df = all_df.loc[all_df[self._PH].isin(phase_numbers)]
        return df.index.min(), df.index.max()

    def parse_days(self, days, ref="last"):
        """Return min(ref, ref + days) and max(ref, ref + days).

        Args:
            days (int): the number of days
            ref (pandas.Timestamp or None): reference date or "first" (the first date of records) or "last" (the last date)

        Note:
            Note that the days clipped with the first and the last dates of records.
        """
        days_n = Validator(days, "days", accept_none=False).int()
        ref_dict = {"first": self._first, "last": self._last}
        ref_date = Validator(ref_dict.get(ref, ref), "ref").date(
            value_range=(self._first, self._last), default=self._last)
        min_date = min(ref_date, ref_date + timedelta(days=days_n))
        max_date = max(ref_date, ref_date + timedelta(days=days_n))
        return max(min_date, self._first), min(max_date, self._last)

    def evaluate(self, date_range=None, metric="RMSLE", display=True, **kwargs):
        """Compare the simulated results and actual records, and evaluate the differences.

        Args:
            date_range (tuple of (str or pandas.Timestamp or None) or None): range of dates to evaluate or None (the first and the last date)
            metric (str): metric to evaluate the difference
            display (bool): whether display figure of comparison or not
            kwargs: keyword arguments of covsirphy.compare_plot()

        Returns:
            float: evaluation score
        """
        variables = [self.CI, self.F, self.R]
        start_date, end_date = Validator(date_range, "date_range").sequence(default=(self._first, self._last), length=2)
        start = Validator(start_date, "date_range[0]").date(value_range=(self._first, self._last), default=self._first)
        end = Validator(end_date, "date_range[1]").date(value_range=(self._first, self._last), default=self._last)
        actual_df = self._df.loc[start:end, variables].dropna(how="any", axis=0)
        sim_df = self.simulate(model_specific=False).loc[start: end, variables].dropna(how="any", axis=0)
        df = actual_df.join(sim_df, how="inner", lsuffix="_actual", rsuffix="_simulated")
        if display:
            compare_plot(df, variables=variables, groups=["actual", "simulated"], **kwargs)
        return Evaluator(actual_df, sim_df).score(metric=metric)

    def start_dates(self):
        """Return the start dates of phases.

        Returns:
            list of [str]: start dates
        """
        df = self._df.reset_index()
        df[self._PH], _ = df[self._PH].factorize()
        return df.groupby(self._PH).first()[self.DATE].tolist()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
try:
    import better_exceptions
    better_exceptions_installed = True
except ImportError:
    better_exceptions_installed = False
# version
from covsirphy.__version__ import __version__
from covsirphy.__citation__ import __citation__
# util
from covsirphy.util.stopwatch import StopWatch
from covsirphy.util.error import deprecate, experimental
from covsirphy.util.error import ExperimentalWarning
from covsirphy.util.error import SubsetNotFoundError, ScenarioNotFoundError
from covsirphy.util.error import PCRIncorrectPreconditionError, NotInteractiveError
from covsirphy.util.error import NotRegisteredError, NotRegisteredMainError, NotRegisteredExtraError
from covsirphy.util.error import UnExpectedReturnValueError, UnExpectedNoneError
from covsirphy.util.error import DBLockedError, NotDBLockedError, NotNoneError, NotEnoughDataError
from covsirphy.util.error import AlreadyCalledError, NotIncludedError, NAFoundError, UnExecutedError, UnExpectedTypeError
from covsirphy.util.error import EmptyError, UnExpectedValueRangeError, UnExpectedValueError, NotSubclassError, UnExpectedLengthError
from covsirphy.util.alias import Alias
from covsirphy.util.filer import save_dataframe
from covsirphy.util.argument import find_args
from covsirphy.util.filer import Filer
from covsirphy.util.evaluator import Evaluator
from covsirphy.util.validator import Validator
from covsirphy.util.term import Term, Word
# visualization
from covsirphy.visualization.vbase import VisualizeBase
from covsirphy._deprecated.colored_map import ColoredMap
from covsirphy._deprecated.japan_map import jpn_map
from covsirphy.visualization.line_plot import LinePlot, line_plot
from covsirphy.visualization.bar_plot import BarPlot, bar_plot
from covsirphy.visualization.compare_plot import ComparePlot, compare_plot
from covsirphy.visualization.scatter_plot import ScatterPlot, scatter_plot
# gis
from covsirphy.gis.gis import GIS
# downloading
from covsirphy.downloading.downloader import DataDownloader
# engineering
from covsirphy.engineering.engineer import DataEngineer
# cleaning
from covsirphy.cleaning.cbase import CleaningBase
from covsirphy._deprecated.jhu_data import JHUData
from covsirphy._deprecated.jhu_complement import JHUDataComplementHandler
from covsirphy._deprecated.country_data import CountryData
from covsirphy.cleaning.japan_data import JapanData
from covsirphy._deprecated.population import PopulationData, Population
from covsirphy.cleaning.pyramid import PopulationPyramidData
from covsirphy._deprecated.oxcgrt import OxCGRTData
from covsirphy._deprecated.pcr_data import PCRData
from covsirphy._deprecated.linelist import LinelistData
from covsirphy._deprecated.vaccine_data import VaccineData
from covsirphy._deprecated.mobility_data import MobilityData
# loading
from covsirphy._deprecated.covid19datahub import COVID19DataHub
from covsirphy._deprecated.dataloader import DataLoader
# trend
from covsirphy._deprecated.trend_detector import TrendDetector, Trend, ChangeFinder
from covsirphy._deprecated.trend_plot import TrendPlot, trend_plot, line_plot_multiple
# ode
from covsirphy._deprecated._mbase import ModelBase
from covsirphy._deprecated._sir import SIR
from covsirphy._deprecated._sird import SIRD
from covsirphy._deprecated._sirf import SIRF
from covsirphy._deprecated._sirfv import SIRFV
from covsirphy._deprecated._sewirf import SEWIRF
from covsirphy._deprecated.ode_handler import ODEHandler
# dynamics
from covsirphy.dynamics.ode import ODEModel
from covsirphy.dynamics.sir import SIRModel
from covsirphy.dynamics.sird import SIRDModel
from covsirphy.dynamics.sirf import SIRFModel
from covsirphy.dynamics.sewirf import SEWIRFModel
from covsirphy.dynamics.dynamics import Dynamics
# simulation
from covsirphy._deprecated.estimator import Estimator, Optimizer
from covsirphy._deprecated.simulator import ODESimulator
# phase
from covsirphy._deprecated.phase_unit import PhaseUnit
from covsirphy._deprecated.phase_series import PhaseSeries
from covsirphy._deprecated.phase_estimator import MPEstimator
# regression
from covsirphy._deprecated.reg_handler import RegressionHandler
# automl
from covsirphy._deprecated.automl_handler import AutoMLHandler
# analysis
from covsirphy._deprecated.example_data import ExampleData
from covsirphy._deprecated.data_handler import DataHandler
from covsirphy._deprecated.param_tracker import ParamTracker
from covsirphy._deprecated.phase_tracker import PhaseTracker
from covsirphy._deprecated.scenario import Scenario
from covsirphy._deprecated.model_validator import ModelValidator
# worldwide
from covsirphy._deprecated.policy import PolicyMeasures
# science
from covsirphy.science.ml import MLEngineer
from covsirphy.science.ode_scenario import ODEScenario


def get_version():
    """
    Return the version number, like CovsirPhy v0.0.0

    Returns:
        str
    """
    return f"CovsirPhy v{__version__}"


def get_citation():
    """
    Return the citation of CovsirPhy

    Returns:
        str
    """
    return __citation__


__all__ = [
    # util-error
    "experimental", "ExperimentalWarning",
    "SubsetNotFoundError", "ScenarioNotFoundError",
    "PCRIncorrectPreconditionError", "NotInteractiveError",
    "NotRegisteredError", "NotRegisteredMainError", "NotRegisteredExtraError",
    "UnExpectedReturnValueError", "DBLockedError", "NotDBLockedError",
    "AlreadyCalledError", "NotIncludedError", "NAFoundError", "UnExecutedError", "UnExpectedTypeError",
    "EmptyError", "UnExpectedValueRangeError", "UnExpectedValueError", "NotSubclassError", "UnExpectedLengthError",
    "Validator", "UnExpectedNoneError", "NotNoneError", "NotEnoughDataError",
    # util
    "StopWatch", "deprecate", "Term", "Filer", "Evaluator", "Alias",
    # visualization
    "VisualizeBase", "LinePlot", "line_plot", "BarPlot", "bar_plot",
    "ComparePlot", "compare_plot", "ScatterPlot", "scatter_plot",
    # gis
    "GIS",
    # downloading
    "DataDownloader",
    # engineer
    "DataEngineer",
    # cleaning
    "CleaningBase", "JapanData", "PopulationPyramidData",
    # dynamics
    "ODEModel", "SIRModel", "SIRDModel", "SIRFModel", "SEWIRFModel", "Dynamics",
    # science
    "MLEngineer", "ODEScenario",
    # Deprecated
    "Population", "Word", "jpn_map", "SIRFV", "line_plot_multiple", "ChangeFinder", "Trend",
    "Optimizer", "save_dataframe", "PolicyMeasures", "ODESimulator", "Estimator", "ParamTracker",
    "PhaseSeries", "PhaseUnit", "MPEstimator", "COVID19DataHub", "LinelistData", "PopulationData",
    "CountryData", "ColoredMap", "ExampleData", "ModelValidator", "find_args",
    "Scenario", "DataHandler", "PhaseTracker",
    "TrendDetector", "TrendPlot", "trend_plot",
    "ModelBase", "SIR", "SIRD", "SIRF", "SEWIRF", "ODEHandler",
    "RegressionHandler",
    "JHUData", "OxCGRTData", "VaccineData", "PCRData", "JHUDataComplementHandler", "MobilityData",
    "DataLoader", "AutoMLHandler",
]

# Show exceptions in better format if used from command line
if not hasattr(sys, "ps1") or not better_exceptions_installed:
    better_exceptions.MAX_LENGTH = None
    better_exceptions.hook()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
import geopandas as gpd
from matplotlib import pyplot as plt
import pandas as pd
from covsirphy.util.error import NotRegisteredError, SubsetNotFoundError
from covsirphy.util.validator import Validator
from covsirphy.util.term import Term
from covsirphy.gis.subset import _SubsetManager
from covsirphy.gis.layer import _LayerAdjuster
from covsirphy.gis.geometry import _Geometry
from covsirphy.gis.choropleth import _ChoroplethMap


class GIS(Term):
    """Class of geographic information system to handle geo-spatial time-series data.

    Args:
        layers (list[str] or None): list of layers of geographic information or None (["ISO3", "Province", "City"])
        country (str or None): layer name of countries or None (countries are not included in the layers)
        date (str): column name of observation dates
        verbose (int): level of verbosity of stdout

    Raises:
        ValueError: @layers has duplicates

    Note:
        Country level data specified with @country will be stored with ISO3 codes.

    Note:
        If @verbose is 0, no descriptions will be shown.
        If @verbose is 1 or larger, details of layer adjustment will be shown.
    """

    def __init__(self, layers=None, country="ISO3", date="Date", verbose=1):
        # Countries will be specified with ISO3 codes and this requires conversion
        self._country = None if country is None else str(country)
        # Location data
        self._layers = Validator(layers or [self._country, self.PROVINCE, self.CITY], "layers").sequence()
        # Date column
        self._date = str(date)
        # Verbosity
        self._verbose = Validator(verbose, "verbose").int(value_range=(0, None))
        # Layer adjuster
        self._adjuster = _LayerAdjuster(
            layers=self._layers, country=self._country, date=self._date, verbose=self._verbose)
        self._un_registered = True

    def all(self, variables=None, errors="raise"):
        """Return all available data.

        Args:
            variables (list[str] or None): list of variables to collect or None (all available variables)
            errors (str): 'raise' or 'coerce'

        Raises:
            NotRegisteredError: No records have been registered yet

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - (pandas.Category): columns defined by covsirphy.GIS(layers)
                    - (pandas.Timestamp): observation dates, column defined by covsirphy.GIS(date)
                    - columns defined by @variables
        """
        if self._un_registered and errors == "raise":
            raise NotRegisteredError("No records have been registered yet.")
        df = self._adjuster.all(variables=variables)
        return df.astype(dict.fromkeys(self._layers, "category"))

    def citations(self, variables=None):
        """
        Return citation list of the secondary data sources.

        Args:
            variables (list[str] or None): list of variables to collect or None (all available variables)

        Returns:
            list[str]: citation list
        """
        return self._adjuster.citations(variables=variables)

    def register(self, data, layers=None, date="Date", variables=None, citations=None, convert_iso3=True, **kwargs):
        """Register new data.

        Args:
            data (pandas.DataFrame): new data
                Index
                    reset index
                Columns
                    - columns defined by @layers
                    - column defined by @date
                    - columns defined by @variables
            layers (list[str] or None): layers of the data or None (as the same as covsirphy.GIS(layer))
            date (str): column name of observation dates of the data
            variables (list[str] or None): list of variables to add or None (all available columns)
            citations (list[str] or str or None): citations of the dataset or None (["my own dataset"])
            convert_iso3 (bool): whether convert country names to ISO3 codes or not
            **kwargs: keyword arguments of pandas.to_datetime() including "dayfirst (bool): whether date format is DD/MM or not"

        Raises:
            ValueError: @data_layers has duplicates

        Returns:
            covsirphy.GIS: self
        """
        self._adjuster.register(
            data=data, layers=layers, date=date, variables=variables, citations=citations,
            convert_iso3=convert_iso3, **Validator(kwargs, "keyword arguments").kwargs(pd.to_datetime))
        self._un_registered = False
        return self

    def layer(self, geo=None, start_date=None, end_date=None, variables=None, errors="raise"):
        """Return the data at the selected layer in the date range.

        Args:
            geo (tuple(list[str] or tuple(str) or str) or str or None): location names to specify the layer or None (the top level)
            start_date (str or None): start date, like 22Jan2020
            end_date (str or None): end date, like 01Feb2020
            variables (list[str] or None): list of variables to add or None (all available columns)
            errors (str): whether raise errors or not, 'raise' or 'coerce'

        Raises:
            TypeError: @geo has un-expected types
            ValueError: the length of @geo is larger than the length of layers
            NotRegisteredError: No records have been registered at the layer yet

        Returns:
            pandas.DataFrame:
                Index:
                    reset index
                Columns
                    - (str): columns defined by covsirphy.GIS(layers)
                    - (pandas.Timestamp): observation dates, column defined by covsirphy.GIS(date)
                    - columns defined by @variables

        Note:
           Note that records with NAs as country names will be always removed.

        Note:
            When `geo=None` or `geo=(None,)`, returns country-level data, assuming we have country/province/city as layers here.

        Note:
            When `geo=("Japan",)` or `geo="Japan"`, returns province-level data in Japan.

        Note:
            When `geo=(["Japan", "UK"],)`, returns province-level data in Japan and UK.

        Note:
            When `geo=("Japan", "Kanagawa")`, returns city-level data in Kanagawa/Japan.

        Note:
            When `geo=("Japan", ["Tokyo", "Kanagawa"])`, returns city-level data in Tokyo/Japan and Kanagawa/Japan.
        """
        # Get all data
        if self._un_registered and errors == "raise":
            raise NotRegisteredError("GIS.register()", details="No records have been registered yet.")
        data = self._adjuster.all(variables=variables)
        # Filter with geo
        geo_converted = self._parse_geo(geo=geo, data=data)
        manager = _SubsetManager(layers=self._layers)
        df = manager.layer(data=data, geo=geo_converted)
        if df.empty and errors == "raise":
            raise NotRegisteredError("GIS.register()", details="No records have been registered at the layer yet.")
        # Filter with date
        series = df[self._date].copy()
        start = Validator(start_date).date(default=series.min())
        end = Validator(end_date).date(default=series.max())
        df = df.loc[(df[self._date] >= start) & (df[self._date] <= end)]
        if df.empty and errors == "raise":
            raise NotRegisteredError(
                "GIS.register()", details=f"No records have been registered at the layer yet from {start_date} to {end_date}.")
        # Get representative records for dates
        df = df.groupby([*self._layers, self._date], dropna=True).first()
        return df.reset_index().convert_dtypes()

    def to_geopandas(self, geo=None, on=None, variables=None, directory=None, natural_earth=None):
        """Add geometry information with GeoJSON file of "Natural Earth" GitHub repository to data.

        Args:
            geo (tuple(list[str] or tuple(str) or str) or str or None): location names to specify the layer or None (the top level)
            on (str or None): the date, like 22Jan2020, or None (the last date of each location)
            variables (list[str] or None): list of variables to add or None (all available columns)
            directory (list[str] or tuple(str) or str): top directory name(s) to save GeoJSON files or None (directory of this this script)
            natural_earth (str or None): title of GeoJSON file (without extension) of "Natural Earth" GitHub repository or None (automatically determined)

        Raises:
            ValueError: country layer is not included in the dataset

        Returns:
            geopandas.GeoDataFrame:
                Index:
                    - reset index
                Columns:
                    - (str): layer focused on with @gis and GIS.layer()
                    - (pandas.Timestamp): observation dates, column defined by covsirphy.GIS(date)
                    - geometry: geometric information

        Note:
            Regarding @geo argument, please refer to covsirphy.GIS.layer().

        Note:
            GeoJSON files are listed in https://github.com/nvkelso/natural-earth-vector/tree/master/geojson
            https://www.naturalearthdata.com/
            https://github.com/nvkelso/natural-earth-vector
            Natural Earth (Free vector and raster map data at naturalearthdata.com, Public Domain)
        """
        if self._country not in self._layers:
            raise ValueError("This cannot be done because country layer is not included in the dataset.")
        df = self.layer(geo=geo, variables=variables)
        if on is None:
            df = df.sort_values(self._date, ascending=True).groupby(self._layers).last().reset_index()
        else:
            df = df.loc[df[self._date] == Validator(on).date()]
        focused_layer = [layer for layer in self._layers if df[layer][df[layer] != self.NA].nunique() > 0][-1]
        geometry = _Geometry(
            data=df, layer=focused_layer, directory=directory or Path(__file__).with_name("Natural_Earth"), verbose=self._verbose)
        iso3 = None if focused_layer == self._country else self._to_iso3(list(df[self._country].unique())[0])
        return geometry.to_geopandas(iso3=iso3, natural_earth=natural_earth).drop(set(self._layers) - {focused_layer}, axis=1)

    def choropleth(self, variable, filename, title="Choropleth map", logscale=True, **kwargs):
        """Create choropleth map.

        Args:
            variable (str): variable name to show
            filename (str or None): filename to save the figure or None (display)
            title (str): title of the map
            logscale (bool): whether convert the value to log10 scale values or not
            kwargs: keyword arguments of the following classes and methods.
                - covsirphy.GIS.to_geopandas() except for @variables,
                - matplotlib.pyplot.savefig(), matplotlib.pyplot.legend(), and
                - pandas.DataFrame.plot()
        """
        v = Validator(kwargs, "keyword arguments")
        gdf = self.to_geopandas(variables=[variable], **v.kwargs(functions=GIS.to_geopandas, default=None))
        focused_layer = [layer for layer in self._layers if layer in gdf.columns][0]
        gdf.rename(columns={focused_layer: "Location", variable: "Variable"}, inplace=True)
        with _ChoroplethMap(filename=filename, **v.kwargs(functions=plt.savefig, default=None)) as cm:
            cm.title = str(title)
            cm.plot(data=gdf, logscale=logscale, **v.kwargs(functions=gpd.GeoDataFrame.plot, default=None))

    def subset(self, geo=None, start_date=None, end_date=None, variables=None, errors="raise"):
        """Return subset of the location and date range.

        Args:
            geo (tuple(list[str] or tuple(str) or str) or str or None): location names to filter or None (total at the top level)
            start_date (str or None): start date, like 22Jan2020
            end_date (str or None): end date, like 01Feb2020
            variables (list[str] or None): list of variables to add or None (all available columns)
            errors (str): whether raise errors or not, 'raise' or 'coerce'

        Raises:
            TypeError: @geo has un-expected types
            ValueError: the length of @geo is larger than the length of layers
            NotRegisteredError: No records have been registered yet
            SubsetNotFoundError: no records were found for the country and @errors is 'raise'

        Returns:
            pandas.DataFrame:
                Index:
                    reset index
                Columns
                    - (pandas.Timestamp): observation dates, column defined by covsirphy.GIS(date)
                    - columns defined by @variables

        Note:
           Note that records with NAs as country names will be always removed.

        Note:
            When `geo=None` or `geo=(None,)`, returns total values of all country-level data, assuming we have country/province/city as layers here.

        Note:
            When `geo=("Japan",)` or `geo="Japan"`, returns country-level data in Japan.

        Note:
            When `geo=(["Japan", "UK"],)`, returns country-level data of Japan and UK.

        Note:
            When `geo=("Japan", "Tokyo")`, returns province-level data of Tokyo/Japan.

        Note:
            When `geo=("Japan", ["Tokyo", "Kanagawa"])`, returns total values of province-level data of Tokyo/Japan and Kanagawa/Japan.

        Note:
            When `geo=("Japan", "Kanagawa", "Yokohama")`, returns city-level data of Yokohama/Kanagawa/Japan.

        Note:
            When `geo=(("Japan", "Kanagawa", ["Yokohama", "Kawasaki"])`, returns total values of city-level data of Yokohama/Kanagawa/Japan and Kawasaki/Kanagawa/Japan.
        """
        # Get all data
        if self._un_registered and errors == "raise":
            raise NotRegisteredError("GIS.register()", details="No records have been registered yet.")
        data = self._adjuster.all(variables=variables)
        # Filter with geo
        geo_converted = self._parse_geo(geo=geo, data=data)
        manager = _SubsetManager(layers=self._layers)
        df = manager.filter(data=data, geo=geo_converted)
        if df.empty and errors == "raise":
            raise SubsetNotFoundError(geo=geo)
        # Filter with date
        series = df[self._date].copy()
        start = Validator(start_date).date(default=series.min())
        end = Validator(end_date).date(default=series.max())
        df = df.loc[(df[self._date] >= start) & (df[self._date] <= end)]
        if df.empty and errors == "raise":
            raise SubsetNotFoundError(geo=geo, start_date=start_date, end_date=end_date)
        # Get representative records for dates
        df = df.groupby([*self._layers, self._date], dropna=True).first().reset_index(level=self._date)
        return df.groupby(self._date, as_index=False).sum().convert_dtypes()

    @classmethod
    def area_name(cls, geo=None):
        """
        Return area name of the geographic information, like 'Japan', 'Tokyo/Japan', 'Japan_UK', 'the world'.

        Args:
            geo (tuple(list[str] or tuple(str) or str) or str or None): location names

        Returns:
            str: area name
        """
        if geo is None or geo[0] is None:
            return "the world"
        names = [
            info if isinstance(info, str) else "_".join(list(info)) for info in ([geo] if isinstance(geo, str) else geo)]
        return cls.SEP.join(names[::-1])

    def _parse_geo(self, geo, data):
        """Parse geographic specifier.

        Args:
            geo (tuple(list[str] or tuple(str) or str) or str or None): location names
            data (pandas.DataFrame):
                Index
                    reset index
                Columns
                    - (str): column defined by @country (of covsirphy.GIS) if @country is not None

        Returns:
            geo (tuple(list[str] or tuple(str) or str) or str or None): parsed location names
        """
        if geo is None:
            return geo
        return [self._info_to_iso3(info, self._layers[i], data) for i, info in enumerate([geo] if isinstance(geo, str) else geo)]

    def _info_to_iso3(self, geo_info, layer, data):
        """Convert a element of geographic specifier to ISO3 code.

        Args:
            geo_info (list[str] or tuple(str) or str or None): element of geographic specifier
            layer (str): layer of geographic information
            data (pandas.DataFrame):
                Index
                    reset index
                Columns
                    - (str): column defined by @country if @country is not None
        """
        if layer != self._country or geo_info is None or set(geo_info).issubset(data[layer].unique()):
            return geo_info
        return self._to_iso3(geo_info)



#!/usr/bin/env python
# -*- coding: utf-8 -*-


from datetime import timedelta
import pandas as pd
from covsirphy.util.error import NAFoundError
from covsirphy.util.validator import Validator
from covsirphy.util.term import Term
from covsirphy.dynamics.ode import ODEModel


class _Simulator(Term):
    """Perform simulation with phase-dependent ODE model.

    Args:
        model (covsirphy.ODEModel): definition of ODE model
        data (pandas.DataFrame): new data to overwrite the current information
            Index
                Date (pandas.Timestamp): Observation date
            Columns
                Phase_ID (int): identification number of phases
                Susceptible (int): the number of susceptible cases
                Infected (int): the number of currently infected cases
                Recovered (int): the number of recovered cases
                Fatal (int): the number of fatal cases
                (numpy.float64): ODE parameter values defined with model.PARAMETERS
    """

    def __init__(self, model, data):
        self._model = Validator(model, "model", accept_none=False).subclass(ODEModel)
        self._df = Validator(data, "data", accept_none=False).dataframe(
            time_index=True, columns=[self._PH, *self._SIRF, *model._PARAMETERS])
        self._first, self._last = data.index.min(), data.index.max()

    def run(self, tau, model_specific=False):
        """Perform simulation with phase-dependent ODE model.

        Args:
            tau (int): tau value [min]
            model_specific (bool): whether convert S, I, F, R to model-specific variables or not

        Raises:
            NAFoundError: ODE parameter values on the start dates of phases are un-set

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    Date (pd.Timestamp): Observation date
                    if @model_specific is False:
                    Susceptible (int): the number of susceptible cases
                    Infected (int): the number of currently infected cases
                    Recovered (int): the number of recovered cases
                    Fatal (int): the number of fatal cases
                    if @model_specific is True, variables defined by model.VARIABLES of covsirphy.Dynamics(model)
        """
        all_df = self._df.copy()
        all_df[self._PH], _ = all_df[self._PH].factorize()
        date_groupby = all_df.reset_index().groupby(self._PH)
        start_dates = date_groupby.first()[self.DATE].sort_values()
        end_dates = date_groupby.last()[self.DATE].sort_values()
        variables, parameters = self._SIRF, self._model._PARAMETERS[:]
        param_df = all_df.ffill().loc[:, parameters]
        # Simulation
        dataframes = []
        for (start, end) in zip(start_dates, end_dates):
            variable_df = all_df.loc[start: end + timedelta(days=1), variables]
            if variable_df.iloc[0].isna().any():
                variable_df = pd.DataFrame(
                    index=pd.date_range(start, end + timedelta(days=1), freq="D"), columns=self._SIRF)
                variable_df.update(self._model.inverse_transform(dataframes[-1]))
            variable_df.index.name = self.DATE
            if param_df.loc[start].isna().any():
                raise NAFoundError(
                    f"ODE parameter values on {start.strftime(self.DATE_FORMAT)}", value=param_df.loc[start].to_dict(),
                    details="Please set values with .register() or .estimate_params()")
            param_dict = param_df.loc[start].to_dict()
            instance = self._model.from_data(data=variable_df.reset_index(), param_dict=param_dict, tau=tau)
            dataframes.append(instance.solve())
        # Combine results of phases
        df = pd.concat(dataframes, axis=0).groupby(level=0).first().loc[self._first: self._last]
        if model_specific:
            return df.reset_index().convert_dtypes()
        df = self._model.inverse_transform(data=df)
        df.index.name = self.DATE
        return df.reset_index().convert_dtypes()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from math import log10, floor
import warnings
import numpy as np
from optuna.exceptions import ExperimentalWarning
import pandas as pd
from sklearn.exceptions import ConvergenceWarning
from covsirphy.util.evaluator import Evaluator
from covsirphy.util.term import Term
from covsirphy._deprecated.reg_pred_actual_plot import _PredActualPlot


class _RegressorBase(Term):
    """
    Basic class to predict parameter values of ODE models.

    Args:
        - X_train (pandas.DataFrame): X for training with time index
        - X_test (pandas.DataFrame): X for test with time index
        - Y_train (pandas.DataFrame): Y for training with time index
        - Y_test (pandas.DataFrame): Y for test with time index
        - X_target (pandas.DataFrame): X for prediction with time index
    """
    warnings.simplefilter("ignore", category=ConvergenceWarning)
    warnings.simplefilter("ignore", category=ExperimentalWarning)
    # Description of regressor
    DESC = ""

    def __init__(self, X_train, X_test, Y_train, Y_test, X_target):
        # Datasets
        self._X_train, self._X_test = X_train.copy(), X_test.copy()
        self._Y_train, self._Y_test = Y_train.copy(), Y_test.copy()
        self._X_target = X_target.copy()
        # Regression model
        self._pipeline = None
        self._param = {}
        # Perform fitting
        self._fit()

    def _fit(self):
        """
        Fit regression model with training dataset, update self._pipeline and self._param.
        This method should be overwritten by child classes.
        """
        raise NotImplementedError

    def score_train(self, metric):
        """
        Calculate score with training dataset.

        Args:
            metric (str): metric name, refer to covsirphy.Evaluator.score()

        Returns:
            float: evaluation score
        """
        pred_train = pd.DataFrame(self._pipeline.predict(self._X_train), columns=self._Y_train.columns)
        return Evaluator(pred_train, self._Y_train, how="all").score(metric=metric)

    def score_test(self, metric):
        """
        Calculate score with test dataset.

        Args:
            metric (str): metric name, refer to covsirphy.Evaluator.score()

        Returns:
            float: evaluation score
        """
        pred_test = pd.DataFrame(self._pipeline.predict(self._X_test), columns=self._Y_test.columns)
        return Evaluator(pred_test, self._Y_test, how="all").score(metric=metric)

    def to_dict(self, metric):
        """
        Calculate score with training/test dataset.

        Args:
            metric (str): metric name, refer to covsirphy.Evaluator.score()

        Returns:
            dict(str, object)
        """
        try:
            param_dict = {
                **self._pipeline.best_params_,
                "intercept": pd.DataFrame(),
                "coef": pd.DataFrame(),
            }
        except AttributeError:
            param_dict = {}
        param_dict.update(self._param)
        return {
            **param_dict,
            "score_metric": metric,
            "score_train": self.score_train(metric=metric),
            "score_test": self.score_test(metric=metric),
            "dataset": {
                "X_train": self._X_train,
                "y_train": self._Y_train,
                "X_test": self._X_test,
                "y_test": self._Y_test,
                "X_target": self._X_target,
            }
        }

    def predict(self):
        """
        Predict parameter values (via Y) with self._pipeline and X_target.

        Returns:
            pandas.DataFrame:
                Index
                    Date (pandas.Timestamp): future dates
                Columns
                    (float): parameter values (4 digits)
        """
        # Predict parameter values
        predicted = self._pipeline.predict(self._X_target)
        df = pd.DataFrame(predicted, index=self._X_target.index, columns=self._Y_train.columns)
        # parameter values: 4 digits
        return df.applymap(lambda x: self._round(x, digits=4))

    def _round(self, value, digits):
        """
        Round off the value to @digits significant digits.

        Args:
            value (float): target value
            digists (int): significant digits

        Return:
            float: rounded value
        """
        return 0 if not value else np.around(value, digits - int(floor(log10(abs(value)))) - 1)

    def _float2str(self, value, to_percentage):
        """
        Convert a float to a string.

        Args:
            value (float): target value
            to_percentage (bool): whether show the value as a percentage or not

        Returns:
            str: the string
        """
        return f"{value:.2%}" if to_percentage else f"{value:.3g}"

    def pred_actual_plot(self, metric, filename=None):
        """
        Create a scatter plot (predicted vs. actual parameter values).

        Args:
            metric (str): metric name, refer to covsirphy.Evaluator.score()
            fileaname (str): filename of the figure or None (display)
        """
        TITLE = f"Predicted vs. actual parameter values\n{self.DESC}"
        PRED, ACTUAL = "Predicted values", "Actual values"
        # Scores
        train_score = self.score_train(metric=metric)
        test_score = self.score_test(metric=metric)
        # Legend
        to_percentage = metric.upper() == "MAPE"
        train_score_str = self._float2str(train_score, to_percentage=to_percentage)
        test_score_str = self._float2str(test_score, to_percentage=to_percentage)
        train_title = f"Training data (n={len(self._X_train)}, {metric}={train_score_str})"
        test_title = f"Test data (n={len(self._X_test)}, {metric}={test_score_str})"
        # Predicted & training
        pred_train = pd.DataFrame(self._pipeline.predict(self._X_train), columns=self._Y_train.columns)
        pred_train["subset"] = train_title
        # Predicted & test
        pred_test = pd.DataFrame(self._pipeline.predict(self._X_test), columns=self._Y_test.columns)
        pred_test["subset"] = test_title
        # Actual & training
        act_train = self._Y_train.copy()
        act_train["subset"] = train_title
        # Actual & test
        act_test = self._Y_train.copy()
        act_test["subset"] = test_title
        # Combine data: index=reset, columns=parameter/subset/Predicted/Actual
        df = pd.concat([pred_train, pred_test], ignore_index=True)
        df = df.melt(id_vars=["subset"], var_name="parameter", value_name=PRED)
        act_df = pd.concat([act_train, act_test], ignore_index=True)
        act_df = act_df.melt(id_vars=["subset"], var_name="parameter", value_name=ACTUAL)
        df.loc[:, ACTUAL] = act_df.loc[:, ACTUAL]
        # Plotting
        with _PredActualPlot(filename=filename) as pa:
            pa.plot(df, x=ACTUAL, y=PRED)
            pa.title = TITLE



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from inspect import signature
import sys
import matplotlib
if not hasattr(sys, "ps1"):
    matplotlib.use("Agg")
from matplotlib import pyplot as plt
from covsirphy.util.error import UnExecutedError
from covsirphy.util.validator import Validator
from covsirphy.util.term import Term

# Style of Matplotlib
plt.style.use("fast")
plt.rcParams["xtick.direction"] = "in"
plt.rcParams["ytick.direction"] = "in"
plt.rcParams["font.size"] = 11.0
plt.rcParams["figure.figsize"] = (9, 6)
plt.rcParams["legend.frameon"] = False


def find_args(func_list, **kwargs):
    """
    Find values of enabled arguments of the function from the keyword arguments.

    Args:
        func_list (list[function] or function): target function
        kwargs: keyword arguments

    Returns:
        dict: dictionary of enabled arguments
    """
    if not isinstance(func_list, list):
        func_list = [func_list]
    enabled_nest = [
        list(signature(func).parameters.keys()) for func in func_list
    ]
    enabled_set = set(sum(enabled_nest, list()))
    enabled_set = enabled_set - {"self", "cls"}
    return {k: v for (k, v) in kwargs.items() if k in enabled_set}


class VisualizeBase(Term):
    """
    Base class for visualization.

    Args:
        filename (str or None): filename to save the figure or None (display)
        bbox_inches (str): bounding box in inches when creating the figure
        kwargs: the other arguments of matplotlib.pyplot.savefig
    """

    def __init__(self, filename=None, bbox_inches="tight", **kwargs):
        self._filename = filename
        self._savefig_dict = {"bbox_inches": bbox_inches, **kwargs}
        # Properties
        self._title = ""
        _, self._ax = plt.subplots(1, 1)

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        # Settings
        if self._title:
            self._ax.title.set_text(self._title)
        # Tight layout
        plt.tight_layout()
        # Display the figure if filename is None after plotting
        if self._filename is None:
            plt.show()
        else:
            # Save the image as a file
            plt.savefig(self._filename, **self._savefig_dict)
            plt.clf()
            plt.close("all")

    @property
    def title(self):
        """
        str: title of the figure
        """
        return self._title

    @title.setter
    def title(self, title):
        self._title = str(title)

    @property
    def ax(self):
        """
        matplotlib.axis: axis
        """
        return self._ax

    @ax.setter
    def ax(self, ax):
        self._ax = Validator(ax, "ax").instance(matplotlib.axes.Axes)

    def plot(self):
        """
        Method for plotting. This will be defined in child classes.

        Raises:
            NotImplementedError: not implemented
        """
        raise NotImplementedError

    def tick_params(self, **kwargs):
        """
        Directly calling matplotlib.pyplot.tick_params,
        change the appearance of ticks, tick labels and grid lines.

        Args:
            kwargs: arguments of matplotlib.pyplot.tick_params
        """
        self._ax.tick_params(**kwargs)

    def legend(self, bbox_to_anchor=(0.5, -0.2), bbox_loc="lower center", ncol=None, **kwargs):
        """
        Set legend.

        Args:
            bbox_to_anchor (tuple(int or float, int or float)): distance of legend and plot
            bbox_loc (str): location of legend
            ncol (int or None): the number of columns that the legend has
            kwargs: keyword arguments of matplotlib.pyplot.legend()
        """
        if not self._variables:
            raise UnExecutedError(".plot()")
        ncol = Validator(
            ncol or (1 if "left" in bbox_loc else len(self._variables)), "ncol").int(value_range=(1, None))
        self._ax.legend(bbox_to_anchor=bbox_to_anchor, loc=bbox_loc, borderaxespad=0, ncol=ncol, **kwargs)
        plt.tight_layout()

    def legend_hide(self):
        """
        Hide legend.
        """
        self._ax.legend().set_visible(False)

    @staticmethod
    def _plot_colors(variables, colormap=None, color_dict=None):
        """
        Create an argument dictionary of colors for Matplotlib.

        Args:
            variables (list[str] or pandas.Index): list of variables to show
            colormap (str, matplotlib colormap object or None): colormap, please refer to https://matplotlib.org/examples/color/colormaps_reference.html
            color_dict (dict[str, str] or None): dictionary of column names (keys) and colors (values)
        """
        # Color
        if color_dict is None:
            return {"colormap": colormap}
        colors = [color_dict.get(col) for col in variables if col in color_dict]
        return {"colormap": colormap, "color": colors}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from covsirphy.util.validator import Validator
from covsirphy.util.term import Term


class _FeatureSelector(Term, BaseEstimator, TransformerMixin):
    """
    Select useful features.
    - Highly correlated features will be replaced with zeros.

    Args:
        corr_threshold (float): lower limit of correlation

    Note:
        We can use this convert in scikit-learn pipeline because this is a child class of
        sklearn.base.BaseEstimator and sklearn.base.TransformerMixin.
    """

    def __init__(self, corr_threshold=0.9):
        self.corr_threshold = Validator(corr_threshold, "corr_threshold").float(value_range=(0, 1))

    @property
    def corr_threshold_(self):
        """
        bool: whether convert the values or not
        """
        return self.corr_threshold

    def fit(self, X, y):
        """
        Return self because fitting is unnecessary.

        Args:
            X (pandas.DataFrame): input X samples
            y (pandas.DataFrame): input y samples

        Returns:
            _FeatureSelector: self
        """
        return self

    def transform(self, X):
        """
        Transform the data.

        Args:
            X (pandas.DataFrame or numpy.ndarray): input samples

        Returns:
            pandas.DataFrame: transformed samples
        """
        df = pd.DataFrame(X)
        # Remove features with high correlation
        corr_matrix = df.corr().abs()
        upper = corr_matrix.where(np.triu(np.ones(corr_matrix.shape), k=1).astype(np.bool))
        dropped_features = [col for col in upper.columns if any(upper[col] > self.corr_threshold)]
        df[dropped_features] = 0
        return df



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd
from covsirphy.util.validator import Validator
from covsirphy.dynamics.sird import SIRDModel


class SIRFModel(SIRDModel):
    """Class of SIR-F model.

    Args:
        date_range (tuple of (str, str)): start date and end date of simulation
        tau (int): tau value [min]
        initial_dict (dict of {str: int}): initial values
            - Susceptible (int): the number of susceptible cases
            - Infected (int): the number of infected cases
            - Fatal (int): the number of fatal cases
            - Recovered (int): the number of recovered cases
        param_dict (dict of {str: float}): non-dimensional parameter values
            - theta: direct fatality probability of un-categorized confirmed cases
            - kappa: non-dimensional mortality rate of infected cases
            - rho: non-dimensional effective contact rate
            - sigma: non-dimensional recovery rate

    Note:
        SIR-F model is original to Covsirphy, https://www.kaggle.com/code/lisphilar/covid-19-data-with-sir-model/notebook
    """
    # Name of ODE model
    _NAME = "SIR-F Model"
    # Non-dimensional parameters
    _PARAMETERS = ["theta", "kappa", "rho", "sigma"]
    # Dimensional parameters
    _DAY_PARAMETERS = ["alpha1 [-]", "1/alpha2 [day]", "1/beta [day]", "1/gamma [day]"]
    # Sample data
    _SAMPLE_DICT = {
        "initial_dict": {SIRDModel.S: 999_000, SIRDModel.CI: 1000, SIRDModel.R: 0, SIRDModel.F: 0},
        "param_dict": {"theta": 0.002, "kappa": 0.005, "rho": 0.2, "sigma": 0.075}
    }

    def __init__(self, date_range, tau, initial_dict, param_dict):
        super().__init__(date_range, tau, initial_dict, param_dict)
        self._theta = Validator(self._param_dict["theta"], "theta", accept_none=False).float(value_range=(0, 1))

    def _discretize(self, t, X):
        """Discretize the ODE.

        Args:
            t (int): discrete time-steps
            X (numpy.array): the current values of the model

        Returns:
            numpy.array: the next values of the model
        """
        n = self._population
        s, i, *_ = X
        dsdt = 0 - self._rho * s * i / n
        drdt = self._sigma * i
        dfdt = self._kappa * i + (0 - dsdt) * self._theta
        didt = 0 - dsdt - drdt - dfdt
        return np.array([dsdt, didt, drdt, dfdt])

    def r0(self):
        """Calculate basic reproduction number.

        Raises:
            ZeroDivisionError: sigma + kappa value was over 0

        Returns:
            float: reproduction number of the ODE model and parameters
        """
        try:
            return round(self._rho * (1 - self._theta) / (self._sigma + self._kappa), 2)
        except ZeroDivisionError:
            raise ZeroDivisionError(
                f"Sigma + kappa must be over 0 to calculate reproduction number with {self._NAME}.") from None

    def dimensional_parameters(self):
        """Calculate dimensional parameter values.

        Raises:
            ZeroDivisionError: either kappa or rho or sigma value was over 0

        Returns:
            dict of {str: int or float}: dictionary of dimensional parameter values
                - "alpha1 [-]" (float): direct fatality probability of un-categorized confirmed cases
                - "1/alpha2 [day]" (int): mortality period of infected cases
                - "1/beta [day]" (int): infection period
                - "1/gamma [day]" (int): recovery period
        """
        try:
            return {
                "alpha1 [-]": round(self._theta, 3),
                "1/alpha2 [day]": round(self._tau / 24 / 60 / self._kappa),
                "1/beta [day]": round(self._tau / 24 / 60 / self._rho),
                "1/gamma [day]": round(self._tau / 24 / 60 / self._sigma)
            }
        except ZeroDivisionError:
            raise ZeroDivisionError(
                f"Kappa, rho and sigma must be over 0 to calculate dimensional parameters with {self._NAME}.") from None

    @classmethod
    def _param_quantile(cls, data, q=0.5):
        """With combinations (X, dX/dt) for X=S, I, R, F, calculate quantile values of ODE parameters.

        Args:
            data (pandas.DataFrame): transformed data with covsirphy.SIRFModel.transform(data=data, tau=tau)
            q (float or array-like): the quantile(s) to compute, value(s) between (0, 1)

        Returns:
            dict of {str: float or pandas.Series}: parameter values at the quantile(s)

        Note:
            We can get approximate parameter values with difference equations as follows.
                - theta -> +0 (i.e. around 0 and not negative)
                - kappa -> (dF/dt) / I when theta -> +0
                - rho = - n * (dS/dt) / S / I
                - sigma = (dR/dt) / I
        """
        df = data.copy()
        periods = round((df.index.max() - df.index.min()) / len(df))
        # Remove negative values and set variables
        df = df.loc[(df[cls.S] > 0) & (df[cls.CI] > 0)]
        n = df.loc[df.index[0], cls._VARIABLES].sum()
        # Calculate parameter values with non-dimensional difference equation
        kappa_series = df[cls.F].diff() / periods / df[cls.CI]
        rho_series = 0 - n * df[cls.S].diff() / periods / df[cls.S] / df[cls.CI]
        sigma_series = df[cls.R].diff() / periods / df[cls.CI]
        # Guess representative values
        return {
            "theta": 0.0 if isinstance(q, float) else pd.Series([0.0, 0.5]).repeat([1, len(q) - 1]),
            "kappa": cls._clip(kappa_series.quantile(q=q), 0, 1),
            "rho": cls._clip(rho_series.quantile(q=q), 0, 1),
            "sigma": cls._clip(sigma_series.quantile(q=q), 0, 1),
        }

    @classmethod
    def sr(cls, data):
        """Return log10(S) and R of model-specific variables for S-R trend analysis.

        Args:
            data (pandas.DataFrame):
                Index
                    Date (pd.Timestamp): Observation date
                Columns
                    Susceptible (int): the number of susceptible cases
                    Infected (int): the number of currently infected cases
                    Recovered (int): the number of recovered cases
                    Fatal (int): the number of fatal cases

        Returns:
            pandas.DataFrame:
                Index
                    Date (pandas.Timestamp): date
                Columns
                    log10(S) (np.float64): common logarithm of Susceptible
                    R (np.int64): Recovered
        """
        Validator(data, "data", accept_none=False).dataframe(time_index=True, columns=cls._SIRF)
        df = data.rename(columns={cls.R: cls._r})
        df[cls._logS] = np.log10(df[cls.S])
        return df.loc[:, [cls._logS, cls._r]].astype({cls._logS: np.float64, cls._r: np.int64})



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
import warnings
import geopandas as gpd
import numpy as np
from mpl_toolkits.axes_grid1 import make_axes_locatable
from covsirphy.util.validator import Validator
from covsirphy.visualization.vbase import VisualizeBase


class _ChoroplethMap(VisualizeBase):
    """Create choropleth map.

    Args:
        filename (str or None): filename to save the figure or None (display)
        kwargs: the other arguments of matplotlib.pyplot.savefig()
    """

    def __init__(self, filename=None, **kwargs):
        super().__init__(filename=filename, **kwargs)
        self._geo_dirpath = Path("input")

    def __enter__(self):
        return super().__enter__()

    def __exit__(self, *exc_info):
        return super().__exit__(*exc_info)

    def plot(self, data, logscale, **kwargs):
        """Set geopandas.GeoDataFrame to create a choropleth map.

        Args:
            data (geopandas.GeoDataFrame):
                Index:
                    - reset index
                Columns:
                    - Location (str): location names
                    - Variable (str): variable to show
                    - geometry: geometric information
            logscale (bool): whether convert the value to log10 scale values or not
            kwargs: arguments of geopandas.GeoDataFrame.plot() except for 'column'
        """
        warnings.filterwarnings("ignore", category=UserWarning)
        df = Validator(data, "data").dataframe(columns=["Location", "Variable", "geometry"])
        df["Variable"] = df["Variable"].astype("float64")
        gdf = gpd.GeoDataFrame(df)
        # Color bar
        divider = make_axes_locatable(self._ax)
        cax = divider.append_axes("bottom", size="5%", pad=0.1)
        # Arguments of plotting with GeoPandas
        plot_kwargs = {
            "legend": True,
            "cmap": "coolwarm",
            "ax": self._ax,
            "cax": cax,
            "missing_kwds": {
                "color": "lightgrey",
                "edgecolor": "white",
                "hatch": "///",
            }
        }
        plot_kwargs.update(kwargs)
        plot_kwargs["legend_kwds"] = {"orientation": "horizontal"}
        # Convert to log10 scale
        if logscale:
            gdf["Variable"] = np.log10(gdf["Variable"] + 1)
            plot_kwargs["legend_kwds"]["label"] = "in log10 scale"
        # Plotting
        if not gdf["Variable"].isna().sum():
            # Missing values are not included
            plot_kwargs.pop("missing_kwds")
        gdf.plot(column="Variable", **plot_kwargs)
        # Remove all ticks
        self._ax.tick_params(labelbottom=False, labelleft=False, left=False, bottom=False)



#!/usr/bin/env python
# -*- coding: utf-8 -*-


from datetime import datetime, timezone, timedelta
from pathlib import Path
import pandas as pd
from covsirphy.util.term import Term
from covsirphy._deprecated.db_cs_japan import _CSJapan
from covsirphy._deprecated.db_covid19dh import _COVID19dh
from covsirphy._deprecated.db_owid import _OWID
from covsirphy._deprecated.db_google import _GoogleOpenData


class _Recommended(Term):
    """
    Download datasets from remote servers.

    Args:
        directory (str or pathlib.Path): directory to save downloaded datasets
        update_interval (int): update interval of downloading dataset
        country (str or None): country name of datasets to downloaded when @update_interval is an integer
        file_dict (dict[str, str] ot None): filename of downloaded CSV files,
            "covid19dh": COVID-19 Data Hub,
            "owid": Our World In Data,
            "google: COVID-19 Open Data by Google Cloud Platform,
            "wbdata": World Bank Open Data,
            "japan": COVID-19 Dataset in Japan.
        verbose (int): level of verbosity when downloading

    Note:
        If @verbose is 0, no descriptions will be shown.
        If @verbose is 1 or larger, URL and database name will be shown.
        If @verbose is 2, detailed citation list will be show, if available.
    """

    def __init__(self, update_interval, country, file_dict, verbose):
        # Update interval of the downloaded files
        self._update_interval = self._ensure_natural_int(update_interval, name="update_interval", include_zero=True)
        # Country names for automated downloading: None (all countries) or list[str]
        self._iso3_code = None if country is None else self._to_iso3(country)[0]
        # Dictionary of filenames to save remote datasets
        self._file_dict = file_dict.copy()
        # Verbosity
        self._verbose = self._ensure_natural_int(verbose, name="verbose", include_zero=True)
        # Column names to identify records
        self._id_cols = [self.ISO3, self.PROVINCE, self.DATE]

    @staticmethod
    def _ensure_natural_int(target, name="number", include_zero=False, none_ok=False):
        """
        Ensure a natural (non-negative) number.

        Args:
            target (int or float or str or None): value to ensure
            name (str): argument name of the value
            include_zero (bool): include 0 or not
            none_ok (bool): None value can be applied or not.

        Returns:
            int: as-is the target

        Note:
            When @target is None and @none_ok is True, None will be returned.
            If the value is a natural number and the type was float or string,
            it will be converted to an integer.
        """
        if target is None and none_ok:
            return None
        s = f"@{name} must be a natural number, but {target} was applied"
        try:
            number = int(target)
        except TypeError as e:
            raise TypeError(f"{s} and not converted to integer.") from e
        if number != target:
            raise ValueError(f"{s}. |{target} - {number}| > 0")
        min_value = 0 if include_zero else 1
        if number < min_value:
            raise ValueError(f"{s}. This value is under {min_value}")
        return number

    def retrieve(self):
        """
        Retrieve datasets from remote servers.

        Returns:
            tuple(pandas/DataFrame, dict[str, str]):
                pandas.DataFrame:
                    Index
                        reset index
                    Columns
                        - Date (str): column name for dates
                        - Country (str): country names (top level administration)
                        - Province (str): province names (2nd level administration)
                        - ISO3 (str): ISO3 codes
                        - Confirmed (numpy.float64): the number of confirmed cases
                        - Fatal (numpy.float64): the number of fatal cases
                        - Recovered (numpy.float64): the number of recovered cases
                        - Population (numpy.int64): population values
                        - Tests (numpy.float64): the number of tests
                        - Product (numpy.float64): vaccine product names
                        - Vaccinations (numpy.float64): cumulative number of vaccinations
                        - Vaccinations_boosters (numpy.float64): cumulative number of booster vaccinations
                        - Vaccinated_once (numpy.float64): cumulative number of people who received at least one vaccine dose
                        - Vaccinated_full (numpy.float64): cumulative number of people who received all doses prescribed by the protocol
                        - School_closing
                        - Workplace_closing
                        - Cancel_events
                        - Gatherings_restrictions
                        - Transport_closing
                        - Stay_home_restrictions
                        - Internal_movement_restrictions
                        - International_movement_restrictions
                        - Information_campaigns
                        - Testing_policy
                        - Contact_tracing
                        - Stringency_index
                        - Mobility_grocery_and_pharmacy: % to baseline in visits (grocery markets, pharmacies etc.)
                        - Mobility_parks: % to baseline in visits (parks etc.)
                        - Mobility_transit_stations: % to baseline in visits (public transport hubs etc.)
                        - Mobility_retail_and_recreation: % to baseline in visits (restaurant, museums etc.)
                        - Mobility_residential: % to baseline in visits (places of residence)
                        - Mobility_workplaces: % to baseline in visits (places of work)


        """
        variables = [
            self.COUNTRY, self.C, self.F, self.R, self.N, self.TESTS,
            self.PRODUCT, self.VAC, self.VAC_BOOSTERS, self.V_ONCE, self.V_FULL,
            *_COVID19dh.OXCGRT_VARS, *_GoogleOpenData.MOBILITY_VARS,
        ]
        all_columns = [*self._id_cols, *variables]
        df = pd.DataFrame(columns=all_columns, index=self._id_cols)
        citation_dict = dict.fromkeys(variables, [])
        # COVID-19 Dataset in Japan
        if self._iso3_code is None or self._iso3_code == "JPN":
            japan_filename = self._file_dict["japan"]
            df, citation_dict = self._add_remote(df, _CSJapan, japan_filename, citation_dict)
        # COVID19 Data Hub
        dh_filename = self._file_dict["covid19dh"]
        df, citation_dict = self._add_remote(df, _COVID19dh, dh_filename, citation_dict)
        # Our World In Data
        owid_filename = self._file_dict["owid"]
        df, citation_dict = self._add_remote(df, _OWID, owid_filename, citation_dict)
        # COVID-19 Open Data by Google Cloud Platform
        google_filename = self._file_dict["google"]
        df, citation_dict = self._add_remote(df, _GoogleOpenData, google_filename, citation_dict)
        # Select country data
        df = df.reset_index()
        if self._iso3_code is not None:
            df = df.loc[df[self.ISO3] == self._iso3_code]
        return (df, citation_dict)

    @staticmethod
    def _last_updated_local(path):
        """
        Return the date last updated of local file/directory.

        Args:
            path (str or pathlibPath): name of the file/directory

        Returns:
            (datetime.datetime): time last updated (UTC)
        """
        m_time = Path(path).stat().st_mtime
        date = datetime.fromtimestamp(m_time)
        return date.astimezone(timezone.utc).replace(tzinfo=None)

    def _download_necessity(self, filename):
        """
        Return whether we need to get the data from remote servers or not,
        comparing the last update of the files.

        Args:
            filename (str): filename of the local file

        Returns:
            (bool): whether we need to get the data from remote servers or not

        Note:
            If the last updated date is unknown, returns True.
            If @update_interval (of _Recommended) hours have passed and the remote file was updated, return True.
        """
        if not Path(filename).exists():
            return True
        date_local = self._last_updated_local(filename)
        time_limit = date_local + timedelta(hours=self._update_interval)
        return datetime.now() > time_limit

    def _add_remote(self, current_df, remote_handler, filename, citation_dict):
        """
        Update null elements of the current database with values in the same date/area in remote dataset.

        Args:
            current_df (pandas.DataFrame): the current database (index: Date, Country, Province)
            remote_handler (covsirphy.loading.db_base._RemoteDatabase): remote database handler class object
            filename (str): filename to save the dataframe retrieved from remote server
            citation_dict (dict[str, list[str]]): dictionary of citation for each variable (column)

        Returns:
            tuple(pandas.DataFrame, list[str]):
                updated database and citations
        """
        df = self._set_date_location(current_df)
        # Get the remote dataset
        force = self._download_necessity(filename)
        handler = remote_handler(filename, self._iso3_code)
        remote_df = self._set_date_location(handler.to_dataframe(force=force, verbose=self._verbose))
        # Update the current database
        df = df.combine_first(remote_df)
        # Update citations
        cite_dict = {k: [*v, handler.CITATION] if k in remote_df else v for (k, v) in citation_dict.items()}
        return (df, cite_dict)

    def _set_date_location(self, data):
        """
        Set date column and location columns.

        Args:
            data (pandas.DataFrame): dataframe to update (itself will be updated)

        Returns:
            pandas.DataFrame:
                Index
                    Date, ISO3, Province
                Columns
                    as-is
        """
        df = data.reset_index().drop("index", axis=1, errors="ignore")
        df[self.DATE] = pd.to_datetime(df[self.DATE])
        df[self.ISO3] = df[self.ISO3].fillna(self.NA)
        if self.COUNTRY in df:
            df[self.COUNTRY] = df[self.COUNTRY].fillna(self.NA)
        df[self.PROVINCE] = df[self.PROVINCE].fillna(self.NA)
        return df.set_index(self._id_cols)

    def covid19dh_citation(self):
        """
        Return the list of primary sources of COVID-19 Data Hub.

        Returns:
            str: the list of primary sources of COVID-19 Data Hub
        """
        dh_handler = _COVID19dh(filename=self._file_dict["covid19dh"], iso3=self._iso3_code)
        return dh_handler.primary()
